"""Module for the Basisset class."""

from __future__ import annotations

import numpy as np

__copyright__ = "Copyright 2024, Molara"

fact2 = [
    1,
    1,
    2,
    3,
    8,
    15,
    48,
    105,
    384,
    945,
    3840,
    10395,
    46080,
    135135,
    645120,
    2027025,
    10321920,
    34459425,
]


class Basisset:
    """Class for either a STO or GTO basisset for each atom in the same order as in molecule.

    GTOs are gaussian type orbitals, STOs are slater type orbitals.
    """

    def __init__(self, basis_type: str = "None") -> None:
        """Initialize the Basisset class.

        :param basis_type: str.
        :return:
        """
        self.basis_type: str = basis_type
        self.orbitals: dict = {}
        self.orbitals_list: list = []

        # self.generate_ijk()

    def generate_orbitals(  # noqa: C901
        self,
        shells: list,
        exponents: list,
        coefficients: list,
        position: np.ndarray,
    ) -> None:
        """Generate the orbitals for the basisset and normalizes the primitive functions.

        :param shells: list of shells
        :param exponents: list of exponents
        :param coefficients: list of coefficients
        :param position: list of positions
        :return:
        """
        i = 0
        si = 1
        pi = 1
        di = 1
        fi = 1
        gi = 1
        orbs = [
            [
                "s",
            ],
            [
                "px",
                "py",
                "pz",
            ],
            [
                "dxx",
                "dyy",
                "dzz",
                "dxy",
                "dxz",
                "dyz",
            ],
            [
                "fxxx",
                "fyyy",
                "fzzz",
                "fxyy",
                "fxxy",
                "fxxz",
                "fxzz",
                "fyzz",
                "fyyz",
                "fxyz",
            ],
            [
                "gxxxx",
                "gyyyy",
                "gzzzz",
                "gxxxy",
                "gxxxz",
                "gyyyx",
                "gyyyz",
                "gzzzx",
                "gzzzy",
                "gxxyy",
                "gxxzz",
                "gyyzz",
                "gxxyz",
                "gyyxz",
                "gzzxy",
            ],
        ]
        for shell in shells:
            ijks: list[list] | np.ndarray = generate_ijks(shell)
            coefficients[i] = np.array(coefficients[i])
            exponents[i] = np.array(exponents[i])
            ijks = np.array(ijks, dtype=int)
            if shell == "s":
                self.orbitals[f"s{si}"] = Orbital(
                    ijks[0],
                    exponents[i],
                    coefficients[i],
                    position,
                )
                si += 1
                i += 1
            elif shell == "p":
                for j, orb in enumerate(orbs[1]):
                    self.orbitals[f"{orb}{pi}"] = Orbital(
                        ijks[j],
                        exponents[i],
                        coefficients[i],
                        position,
                    )
                pi += 1
                i += 1
            elif shell == "d":
                for j, orb in enumerate(orbs[2]):
                    self.orbitals[f"{orb}{di}"] = Orbital(
                        ijks[j],
                        exponents[i],
                        coefficients[i],
                        position,
                    )
                di += 1
                i += 1
            elif shell == "f":
                for j, orb in enumerate(orbs[3]):
                    self.orbitals[f"{orb}{fi}"] = Orbital(
                        ijks[j],
                        exponents[i],
                        coefficients[i],
                        position,
                    )
                fi += 1
                i += 1
            elif shell == "g":
                for j, orb in enumerate(orbs[4]):
                    self.orbitals[f"{orb}{gi}"] = Orbital(
                        ijks[j],
                        exponents[i],
                        coefficients[i],
                        position,
                    )
                gi += 1
                i += 1
            else:
                msg = f"The shell {shell} type is not supported."
                raise TypeError(msg)
            self.orbitals_list = list(self.orbitals.values())


class Orbital:
    """Class to store either an STO or GTO."""

    def __init__(
        self,
        ijk: np.ndarray,
        exponents: np.ndarray,
        coefficients: np.ndarray,
        position: np.ndarray,
    ) -> None:
        """Initialize the orbital class.

        :param ijk: list of ijk values
        :param exponents: list of exponents
        :param coefficients: list of coefficients
        :param position: position of the orbital
        :return:
        """
        self.ijk = ijk
        self.exponents = exponents
        self.norms = calculate_normalization_primitive_gtos(ijk, exponents)
        self.coefficients = coefficients * calculate_normalization_contracted_gtos(
            ijk,
            exponents,
            coefficients,
            self.norms,
        )
        self.position = position


def calculate_normalization_primitive_gtos(
    ijk: np.ndarray,
    exponents: np.ndarray,
) -> np.ndarray:
    """Normalize the primitive gaussians.

    :param ijk: list of ijk values
    :param exponents: list of exponents
    :return: normalization factor
    """
    i = ijk[0]
    j = ijk[1]
    k = ijk[2]
    m = i + j + k

    fi = 1 if i == 0 else fact2[2 * i - 1]
    fj = 1 if j == 0 else fact2[2 * j - 1]
    fk = 1 if k == 0 else fact2[2 * k - 1]

    return np.array(
        [
            np.sqrt(
                (2 ** (2 * m + 1.5) * exponent ** (m + 1.5)) / (fi * fj * fk * np.pi**1.5),
            )
            for exponent in exponents
        ],
    )


def calculate_normalization_contracted_gtos(
    ijk: np.ndarray,
    exponents: np.ndarray,
    coefficients: np.ndarray,
    norms: np.ndarray,
) -> float:
    """Normalize the contracted gaussians.

    :param ijk: list of ijk values
    :param exponents: list of exponents
    :param coefficients: list of coefficients
    :param norms: list of normalization factors
    :return: normalization factor
    """
    i = ijk[0]
    j = ijk[1]
    k = ijk[2]
    m = i + j + k

    fi = 1 if i == 0 else fact2[2 * i - 1]
    fj = 1 if j == 0 else fact2[2 * j - 1]
    fk = 1 if k == 0 else fact2[2 * k - 1]

    prefactor = (np.pi**1.5 * fi * fj * fk) / (2 ** (m))
    n = 0.0

    for ia in range(len(exponents)):
        for ib in range(len(exponents)):
            n += (coefficients[ia] * coefficients[ib] * norms[ia] * norms[ib]) / (exponents[ia] + exponents[ib]) ** (
                m + 1.5
            )
    n = n * prefactor

    return n ** (-0.5)


def generate_ijks(shell: str) -> list[list]:
    """Generate the ijk values for the shells.

    :param shell: name of the atomic orbital shell (s/p/d/f)
    :return:
    """
    if shell == "s":
        return [[0, 0, 0]]
    if shell == "p":
        return [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    if shell == "d":
        return [[2, 0, 0], [0, 2, 0], [0, 0, 2], [1, 1, 0], [1, 0, 1], [0, 1, 1]]
    if shell == "f":
        return [
            [3, 0, 0],
            [0, 3, 0],
            [0, 0, 3],
            [1, 2, 0],
            [2, 1, 0],
            [2, 0, 1],
            [1, 0, 2],
            [0, 1, 2],
            [0, 2, 1],
            [1, 1, 1],
        ]
    if shell == "g":
        return [
            [4, 0, 0],
            [0, 4, 0],
            [0, 0, 4],
            [3, 1, 0],
            [3, 0, 1],
            [1, 3, 0],
            [0, 3, 1],
            [1, 0, 3],
            [0, 1, 3],
            [2, 2, 0],
            [2, 0, 2],
            [0, 2, 2],
            [2, 1, 1],
            [1, 2, 1],
            [1, 1, 2],
        ]
    msg = f"The shell {shell} type is not supported."
    raise TypeError(msg)



"""This module contains the Cylinder and Cylinders classes.

They are used to create cylinders and multiple cylinders of the same color, respectively.
"""

cimport numpy as npc
import numpy as np
from molara.tools.mathtools import norm_float


class Cylinder:
    """Creates a Cylinder object, containing its vertices and indices.

    :param color: Color of the cylinder.
    :param subdivisions: Number of subdivisions of the cylinder.
    """

    def __init__(self, subdivisions: int) -> None:
        """Creates a Cylinder object, containing its vertices and indices."""
        self.subdivisions = subdivisions
        vertices, indices = generate_cylinder(self.subdivisions)
        self.vertices = vertices
        self.indices = indices


def generate_cylinder(
    subdivisions: int,
) -> tuple[np.ndarray, np.ndarray]:
    """Calculates the vertices and indices of a cylinder for a given number of subdivisions.

    :param color: Color of the cylinder.
    :param subdivisions: Number of subdivisions of the cylinder.
    :returns:
        - **vertices** (numpy.array of numpy.float32) - Vertices in the following order x,y,z,nx,ny,nz,..., where\
         xyz are the cartesian coordinates, and nxnynz are the components of the normal\
          vector.
        - **indices** (numpy.array of numpy.uint32) - Gives the connectivity of the vertices.
    """
    vertices = []
    indices = []

    height = 1.0
    radius = 1.0
    vertices.extend([0.0, -height / 2, 0.0, 0.0, -1, 0.0])
    vertices.extend([0.0, height / 2, 0.0, 0.0, 1, 0.0])
    for i in range(subdivisions):
        # vertices
        theta = 2 * np.pi * i / subdivisions
        x = radius * np.cos(theta)
        y = -height / 2
        z = radius * np.sin(theta)
        normal = np.array([x, 0, z],dtype=np.float32)
        normal /= norm_float(normal)
        vertices.extend([x, y, z, 0, -1, 0])
        vertices.extend(
            [x, y, z, normal[0], normal[1], normal[2]],
        )

        vertices.extend([x, y + height, z, 0, 1, 0])
        vertices.extend(
            [
                x,
                y + height,
                z,
                normal[0],
                normal[1],
                normal[2],
            ],
        )

        # indices
        if i == subdivisions - 1:
            # top
            indices.extend([0, 2 + 4 * i, 2])
            # bottom
            indices.extend([1, 4 + 4 * i, 4])
            # side
            indices.extend([3 + 4 * i, 3, 5])
            indices.extend([5 + 4 * i, 3 + 4 * i, 5])

        else:
            # top
            indices.extend([0, 2 + 4 * i, 6 + 4 * i])
            # bottom
            indices.extend([1, 4 + 4 * i, 8 + 4 * i])
            # side
            indices.extend([3 + 4 * i, 7 + 4 * i, 5 + 4 * i])
            indices.extend([5 + 4 * i, 7 + 4 * i, 9 + 4 * i])

    return np.array(vertices, dtype=np.float32), np.array(indices, dtype=np.uint32)


cpdef calculate_cylinder_model_matrix(
    npc.ndarray[float, ndim=1] position,
    float radius,
    float length,
    npc.ndarray[float, ndim=1] direction,
):
    """Calculates the model matrix for a cylinder.

    :param position: Position of the center of the cylinder.
    :param radius: Radius of the cylinder.
    :param length: Length of the cylinder.
    :param direction: Direction of the cylinder, does not need to be normalized.
    """
    cdef npc.ndarray[float, ndim=1] rotation_axis = np.empty(3, dtype=np.float32)
    cdef float rotation_angle, x, y, z, c, s, t
    cdef npc.ndarray[float, ndim=2] rotation_scale_matrix
    cdef npc.ndarray[float, ndim=2] scale_matrix
    cdef npc.ndarray[float, ndim=2] rotation_matrix
    cdef npc.ndarray[float, ndim=2] translation_matrix
    cdef float[3] y_axis = np.array([0, 1, 0], dtype=np.float32)
    cdef float direction_norm = norm_float(direction)
    direction = direction / direction_norm
    cdef float dot = np.dot(direction, y_axis)
    if abs(dot) != 1:
        rotation_axis = np.array([
        y_axis[1] * direction[2] - y_axis[2] * direction[1],
        y_axis[2] * direction[0] - y_axis[0] * direction[2],
        y_axis[0] * direction[1] - y_axis[1] * direction[0]
        ], dtype=np.float32)
        # Calculate the angle to rotate the cylinder to the correct orientation.
        rotation_angle = np.arccos(
            np.clip(
                dot,
                -1,
                1,
            ),
        )
    else:
        rotation_axis = np.array([0, 0, 1], dtype=np.float32)
        rotation_angle = 0
    translation_matrix = np.array([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]], dtype=np.float32)
    translation_matrix[3, 0:3] = position

    rotation_matrix = np.array([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]], dtype=np.float32)

    rotation_axis = rotation_axis / norm_float(rotation_axis)
    x, y, z = rotation_axis
    c = np.cos(rotation_angle)
    s = np.sin(rotation_angle)
    t = 1 - c

    rotation_matrix[:3, :3] = np.array([
        [t*x*x + c, t*x*y + s*z, t*x*z - s*y],
        [t*x*y - s*z, t*y*y + c, t*y*z + s*x],
        [t*x*z + s*y, t*y*z - s*x, t*z*z + c]]
    , dtype=np.float32)

    cdef float[3] scale = [radius, radius, radius]
    scale[1] = length
    scale_matrix =np.array([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]], dtype=np.float32)
    scale_matrix[0, 0] = scale[0]
    scale_matrix[1, 1] = scale[1]
    scale_matrix[2, 2] = scale[2]
    rotation_scale_matrix = scale_matrix @ rotation_matrix
    return np.array([rotation_scale_matrix @ translation_matrix], dtype=np.float32)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QMainWindow, QMenu,
    QMenuBar, QSizePolicy, QStatusBar, QWidget)

from molara.gui.structure_widget import StructureWidget

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(800, 600)
        self.quit = QAction(MainWindow)
        self.quit.setObjectName(u"quit")
        self.action_xyz = QAction(MainWindow)
        self.action_xyz.setObjectName(u"action_xyz")
        self.actionCenter_Molecule = QAction(MainWindow)
        self.actionCenter_Molecule.setObjectName(u"actionCenter_Molecule")
        self.actionReset_View = QAction(MainWindow)
        self.actionReset_View.setObjectName(u"actionReset_View")
        self.actionto_x_axis = QAction(MainWindow)
        self.actionto_x_axis.setObjectName(u"actionto_x_axis")
        self.actionto_y_axis = QAction(MainWindow)
        self.actionto_y_axis.setObjectName(u"actionto_y_axis")
        self.actionto_z_axis = QAction(MainWindow)
        self.actionto_z_axis.setObjectName(u"actionto_z_axis")
        self.actionDraw_Axes = QAction(MainWindow)
        self.actionDraw_Axes.setObjectName(u"actionDraw_Axes")
        self.actionCreate_Lattice = QAction(MainWindow)
        self.actionCreate_Lattice.setObjectName(u"actionCreate_Lattice")
        self.actionRead_POSCAR = QAction(MainWindow)
        self.actionRead_POSCAR.setObjectName(u"actionRead_POSCAR")
        self.action_coord = QAction(MainWindow)
        self.action_coord.setObjectName(u"action_coord")
        self.actionOpen_Trajectory_Dialog = QAction(MainWindow)
        self.actionOpen_Trajectory_Dialog.setObjectName(u"actionOpen_Trajectory_Dialog")
        self.actionImport = QAction(MainWindow)
        self.actionImport.setObjectName(u"actionImport")
        self.actionToggle_Bonds = QAction(MainWindow)
        self.actionToggle_Bonds.setObjectName(u"actionToggle_Bonds")
        self.actionMeasure = QAction(MainWindow)
        self.actionMeasure.setObjectName(u"actionMeasure")
        self.actionSupercell = QAction(MainWindow)
        self.actionSupercell.setObjectName(u"actionSupercell")
        self.actionBuilder = QAction(MainWindow)
        self.actionBuilder.setObjectName(u"actionBuilder")
        self.actionExport = QAction(MainWindow)
        self.actionExport.setObjectName(u"actionExport")
        self.actionExport_Snapshot = QAction(MainWindow)
        self.actionExport_Snapshot.setObjectName(u"actionExport_Snapshot")
        self.actionToggle_UnitCellBoundaries = QAction(MainWindow)
        self.actionToggle_UnitCellBoundaries.setObjectName(u"actionToggle_UnitCellBoundaries")
        self.actionToggle_Projection = QAction(MainWindow)
        self.actionToggle_Projection.setObjectName(u"actionToggle_Projection")
        self.actionExport_CameraSettings = QAction(MainWindow)
        self.actionExport_CameraSettings.setObjectName(u"actionExport_CameraSettings")
        self.actionImport_CameraSettings = QAction(MainWindow)
        self.actionImport_CameraSettings.setObjectName(u"actionImport_CameraSettings")
        self.actionOpen_Structure_Customizer = QAction(MainWindow)
        self.actionOpen_Structure_Customizer.setObjectName(u"actionOpen_Structure_Customizer")
        self.actionNumber_Atoms = QAction(MainWindow)
        self.actionNumber_Atoms.setObjectName(u"actionNumber_Atoms")
        self.actionToggle_Axes = QAction(MainWindow)
        self.actionToggle_Axes.setObjectName(u"actionToggle_Axes")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.openGLWidget = StructureWidget(self.centralwidget)
        self.openGLWidget.setObjectName(u"openGLWidget")

        self.gridLayout.addWidget(self.openGLWidget, 0, 0, 1, 1)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 800, 37))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuImport = QMenu(self.menuFile)
        self.menuImport.setObjectName(u"menuImport")
        self.menuExport = QMenu(self.menuFile)
        self.menuExport.setObjectName(u"menuExport")
        self.menuEdit = QMenu(self.menubar)
        self.menuEdit.setObjectName(u"menuEdit")
        self.menuRotate = QMenu(self.menuEdit)
        self.menuRotate.setObjectName(u"menuRotate")
        self.menuCrystal = QMenu(self.menubar)
        self.menuCrystal.setObjectName(u"menuCrystal")
        self.menuTools = QMenu(self.menubar)
        self.menuTools.setObjectName(u"menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuCrystal.menuAction())
        self.menuFile.addAction(self.menuImport.menuAction())
        self.menuFile.addAction(self.menuExport.menuAction())
        self.menuFile.addAction(self.quit)
        self.menuImport.addAction(self.actionImport)
        self.menuImport.addAction(self.actionImport_CameraSettings)
        self.menuImport.addAction(self.actionRead_POSCAR)
        self.menuExport.addAction(self.actionExport)
        self.menuExport.addAction(self.actionExport_Snapshot)
        self.menuExport.addAction(self.actionExport_CameraSettings)
        self.menuEdit.addAction(self.actionCenter_Molecule)
        self.menuEdit.addAction(self.actionReset_View)
        self.menuEdit.addAction(self.menuRotate.menuAction())
        self.menuEdit.addAction(self.actionToggle_Projection)
        self.menuEdit.addAction(self.actionToggle_Axes)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionOpen_Structure_Customizer)
        self.menuRotate.addAction(self.actionto_x_axis)
        self.menuRotate.addAction(self.actionto_y_axis)
        self.menuRotate.addAction(self.actionto_z_axis)
        self.menuCrystal.addAction(self.actionCreate_Lattice)
        self.menuCrystal.addAction(self.actionSupercell)
        self.menuCrystal.addAction(self.actionToggle_UnitCellBoundaries)
        self.menuTools.addAction(self.actionMeasure)
        self.menuTools.addAction(self.actionBuilder)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.quit.setText(QCoreApplication.translate("MainWindow", u"Quit", None))
        self.action_xyz.setText(QCoreApplication.translate("MainWindow", u".xyz", None))
        self.actionCenter_Molecule.setText(QCoreApplication.translate("MainWindow", u"Center Molecule", None))
        self.actionReset_View.setText(QCoreApplication.translate("MainWindow", u"Reset View", None))
        self.actionto_x_axis.setText(QCoreApplication.translate("MainWindow", u"to x axis", None))
        self.actionto_y_axis.setText(QCoreApplication.translate("MainWindow", u"to y axis", None))
        self.actionto_z_axis.setText(QCoreApplication.translate("MainWindow", u"to z axis", None))
        self.actionDraw_Axes.setText(QCoreApplication.translate("MainWindow", u"Toggle Axes", None))
        self.actionCreate_Lattice.setText(QCoreApplication.translate("MainWindow", u"Create Lattice", None))
        self.actionRead_POSCAR.setText(QCoreApplication.translate("MainWindow", u"Import POSCAR", None))
        self.action_coord.setText(QCoreApplication.translate("MainWindow", u"coord", None))
        self.actionOpen_Trajectory_Dialog.setText(QCoreApplication.translate("MainWindow", u"Open Trajectory Dialog", None))
        self.actionImport.setText(QCoreApplication.translate("MainWindow", u"Import Structure File", None))
        self.actionToggle_Bonds.setText(QCoreApplication.translate("MainWindow", u"Toggle Bonds", None))
        self.actionMeasure.setText(QCoreApplication.translate("MainWindow", u"Measure", None))
        self.actionSupercell.setText(QCoreApplication.translate("MainWindow", u"Supercell Dimensions", None))
        self.actionBuilder.setText(QCoreApplication.translate("MainWindow", u"Z-Matrix Builder", None))
        self.actionExport.setText(QCoreApplication.translate("MainWindow", u"Export Structure File", None))
        self.actionExport_Snapshot.setText(QCoreApplication.translate("MainWindow", u"Export Snapshot", None))
        self.actionToggle_UnitCellBoundaries.setText(QCoreApplication.translate("MainWindow", u"Toggle Unit Cell Boundaries", None))
        self.actionToggle_Projection.setText(QCoreApplication.translate("MainWindow", u"Toggle Projection (Perspective/Ortho)", None))
        self.actionExport_CameraSettings.setText(QCoreApplication.translate("MainWindow", u"Export Camera Settings", None))
        self.actionImport_CameraSettings.setText(QCoreApplication.translate("MainWindow", u"Import Camera Settings", None))
        self.actionOpen_Structure_Customizer.setText(QCoreApplication.translate("MainWindow", u"Structure Customizer", None))
        self.actionNumber_Atoms.setText(QCoreApplication.translate("MainWindow", u"Number Atoms", None))
        self.actionToggle_Axes.setText(QCoreApplication.translate("MainWindow", u"Show Axes", None))
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"File", None))
        self.menuImport.setTitle(QCoreApplication.translate("MainWindow", u"Import", None))
        self.menuExport.setTitle(QCoreApplication.translate("MainWindow", u"Export", None))
        self.menuEdit.setTitle(QCoreApplication.translate("MainWindow", u"View", None))
        self.menuRotate.setTitle(QCoreApplication.translate("MainWindow", u"Set View", None))
        self.menuCrystal.setTitle(QCoreApplication.translate("MainWindow", u"Crystal", None))
        self.menuTools.setTitle(QCoreApplication.translate("MainWindow", u"Tools", None))
    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'export_image_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QDialog,
    QDialogButtonBox, QDoubleSpinBox, QLabel, QLineEdit,
    QPushButton, QSizePolicy, QSpinBox, QTabWidget,
    QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(402, 266)
        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(200, 220, 181, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.tabWidget = QTabWidget(Dialog)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tabWidget.setGeometry(QRect(20, 10, 361, 191))
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.label_2 = QLabel(self.tab)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(10, 70, 91, 16))
        self.heightSpinBox = QDoubleSpinBox(self.tab)
        self.heightSpinBox.setObjectName(u"heightSpinBox")
        self.heightSpinBox.setGeometry(QRect(120, 60, 62, 26))
        self.heightSpinBox.setDecimals(0)
        self.heightSpinBox.setMinimum(1.000000000000000)
        self.heightSpinBox.setMaximum(50000.000000000000000)
        self.checkBox = QCheckBox(self.tab)
        self.checkBox.setObjectName(u"checkBox")
        self.checkBox.setGeometry(QRect(200, 40, 131, 21))
        self.widthSpinBox = QDoubleSpinBox(self.tab)
        self.widthSpinBox.setObjectName(u"widthSpinBox")
        self.widthSpinBox.setGeometry(QRect(120, 20, 62, 26))
        self.widthSpinBox.setDecimals(0)
        self.widthSpinBox.setMinimum(1.000000000000000)
        self.widthSpinBox.setMaximum(50000.000000000000000)
        self.label = QLabel(self.tab)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(10, 20, 91, 16))
        self.filenameInput = QLineEdit(self.tab)
        self.filenameInput.setObjectName(u"filenameInput")
        self.filenameInput.setGeometry(QRect(10, 110, 241, 28))
        self.pushButton = QPushButton(self.tab)
        self.pushButton.setObjectName(u"pushButton")
        self.pushButton.setGeometry(QRect(260, 110, 90, 28))
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.sphereSubdivisionsSpinBox = QSpinBox(self.tab_2)
        self.sphereSubdivisionsSpinBox.setObjectName(u"sphereSubdivisionsSpinBox")
        self.sphereSubdivisionsSpinBox.setGeometry(QRect(160, 20, 42, 26))
        self.cylinderSubdivisionsSpinBox = QSpinBox(self.tab_2)
        self.cylinderSubdivisionsSpinBox.setObjectName(u"cylinderSubdivisionsSpinBox")
        self.cylinderSubdivisionsSpinBox.setGeometry(QRect(160, 70, 42, 26))
        self.label_3 = QLabel(self.tab_2)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setGeometry(QRect(20, 20, 121, 16))
        self.label_4 = QLabel(self.tab_2)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setGeometry(QRect(20, 80, 131, 16))
        self.tabWidget.addTab(self.tab_2, "")
        QWidget.setTabOrder(self.tabWidget, self.widthSpinBox)
        QWidget.setTabOrder(self.widthSpinBox, self.heightSpinBox)
        QWidget.setTabOrder(self.heightSpinBox, self.checkBox)
        QWidget.setTabOrder(self.checkBox, self.filenameInput)
        QWidget.setTabOrder(self.filenameInput, self.pushButton)
        QWidget.setTabOrder(self.pushButton, self.sphereSubdivisionsSpinBox)
        QWidget.setTabOrder(self.sphereSubdivisionsSpinBox, self.cylinderSubdivisionsSpinBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Export Snapshot", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Height (pixels)", None))
        self.checkBox.setText(QCoreApplication.translate("Dialog", u"Keep aspect ratio", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Width (pixels)", None))
        self.pushButton.setText(QCoreApplication.translate("Dialog", u"Change file", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("Dialog", u"Export Options", None))
        self.label_3.setText(QCoreApplication.translate("Dialog", u"Sphere Subdivisions", None))
        self.label_4.setText(QCoreApplication.translate("Dialog", u"Cylinder Subdivisions", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("Dialog", u"Rendering Options", None))
    # retranslateUi



"""The main entry point for the application."""

from __future__ import annotations

import signal
import sys
from pathlib import Path

from PySide6.QtGui import QIcon, QSurfaceFormat
from PySide6.QtWidgets import QApplication

# Important:
# You need to run the following command to generate the ui_form.py file
#     pyside6-uic form.ui -o ui_form.py, or
#     pyside2-uic form.ui -o ui_form.py
from molara.gui.main_window import MainWindow

__copyright__ = "Copyright 2024, Molara"


def main(test: bool = False) -> None:
    """Run the application.

    :param test: True if the application is run in test mode, False otherwise
    """
    _format = QSurfaceFormat()
    _format.setVersion(3, 3)
    _format.setSamples(4)
    _format.setProfile(QSurfaceFormat.CoreProfile)  # type: ignore[attr-defined]
    QSurfaceFormat.setDefaultFormat(_format)

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(f"{Path(__file__).resolve().parent}/MolaraLogo.png"))

    widget = MainWindow()

    widget.setWindowTitle("Molara")

    if not test:
        widget.show()

        if len(sys.argv) > 1:
            widget.show_init_xyz()

        sys.exit(app.exec())


if __name__ == "__main__":
    main(test=False)



"""Dialog for manipulating appearance of structures."""

from __future__ import annotations

import json
import shutil
from os import listdir
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from PySide6.QtWidgets import (
    QDialog,
    QMainWindow,
)

from molara.gui.ui_structure_customizer import Ui_structure_customizer
from molara.rendering.atom_labels import init_atom_number

if TYPE_CHECKING:
    from molara.structure.crystal import Crystal
    from molara.structure.molecule import Molecule
    from molara.structure.structure import Structure

__copyright__ = "Copyright 2024, Molara"


class StructureCustomizerDialog(QDialog):
    """Dialog for manipulating appearance of trajectories."""

    def __init__(self, parent: QMainWindow = None) -> None:
        """Initialize the trajectory dialog.

        :param parent: parent widget (main window)
        """
        super().__init__(
            parent,
        )  # main window widget is passed as a parent, so dialog is closed if main window is closed.
        self.ui = Ui_structure_customizer()
        self.ui.setupUi(self)

        self.src_path = Path(__file__).parent.parent
        self.home_path = Path("~/.molara").expanduser()
        self.save_names: list = []
        self.update_settings_box()

        self.stick_mode = False
        self.bonds = True
        self.numbers = False
        self.max_atoms_for_numbers = 999
        self.atom_indices_arrays: tuple[np.ndarray, np.ndarray, np.ndarray] = (np.zeros(1), np.zeros(1), np.zeros(1))

        self.ui.ballSizeSpinBox.setValue(1.0)
        self.ui.stickSizeSpinBox.setValue(1.0)
        self.ui.toggleBondsButton.setText("Hide Bonds")
        self.ui.toggleNumbersButton.setText("Show Indices")
        self.ui.colorSchemeSelect.addItems(["Jmol", "CPK"])

        self.ui.viewModeButton.clicked.connect(self.toggle_stick_mode)
        self.ui.toggleBondsButton.clicked.connect(self.toggle_bonds)
        self.ui.toggleNumbersButton.clicked.connect(self.toggle_numbers)
        self.ui.saveButton.clicked.connect(self.save_settings)
        self.ui.loadButton.clicked.connect(self.load_settings)
        self.ui.deleteButton.clicked.connect(self.delete_settings)

        self.ui.ballSizeSpinBox.valueChanged.connect(self.apply_changes)
        self.ui.stickSizeSpinBox.valueChanged.connect(self.apply_changes)
        self.ui.indexSizeSpinBox.valueChanged.connect(self.apply_changes)
        self.ui.colorSchemeSelect.currentIndexChanged.connect(self.apply_changes)

        self.load_default_settings()

    def load_default_settings(self) -> None:
        """Load the default settings."""
        self.load_settings("Default")

    def update_settings_box(self) -> None:
        """Update the settings box."""
        if not self.home_path.joinpath("settings/structure").exists():
            Path(f"{self.home_path}/settings/structure").mkdir(parents=True, exist_ok=True)
        if not self.home_path.joinpath("settings/structure/Default.json").exists():
            shutil.copy(
                f"{self.src_path}/settings/structure/Default.json",
                f"{self.home_path}/settings/structure/Default.json",
            )
        save_files = [
            f
            for f in listdir(f"{self.home_path}/settings/structure")
            if Path.is_file(self.home_path / "settings" / "structure" / f)
        ]
        self.save_names = [f.split(".")[0] for f in save_files]
        self.save_names.sort()
        self.ui.loadSelect.clear()
        self.ui.loadSelect.addItems(self.save_names)

    def delete_settings(self) -> None:
        """Delete the selected settings."""
        save_name = self.ui.loadSelect.currentText()
        if save_name == "Default":
            return
        settings_file = f"{self.home_path}/settings/structure/{save_name}.json"
        Path(settings_file).unlink()
        self.update_settings_box()

    def load_settings(self, save_name: str = "") -> None:
        """Load the settings from a save file.

        :param save_name: name of the save file
        """
        if not save_name:
            save_name = self.ui.loadSelect.currentText()
        settings_file = f"{self.home_path}/settings/structure/{save_name}.json"
        with open(settings_file) as f:
            settings = json.load(f)
        self.load_settings_dict(settings)
        if self.parent().structure_widget.structures:
            self.apply_changes()

    def create_settings_dict(self) -> dict:
        """Create a dictionary of the settings."""
        return {
            "stick_mode": bool(self.stick_mode),
            "bonds": bool(self.bonds),
            "ball_size": float(self.ui.ballSizeSpinBox.value()),
            "stick_size": float(self.ui.stickSizeSpinBox.value()),
            "atom_numbers": bool(self.numbers),
            "atom_numbers_size": float(self.ui.indexSizeSpinBox.value()),
            "color_scheme": self.ui.colorSchemeSelect.currentText(),
        }

    def load_settings_dict(self, settings: dict) -> None:
        """Load the settings from a dictionary.

        settings: dict: dictionary of settings
        """
        self.stick_mode = settings["stick_mode"]
        if self.stick_mode:
            self.ui.viewModeButton.setText("Ball Mode")
            self.ui.stickSizeSpinBox.setEnabled(True)
            self.ui.ballSizeSpinBox.setEnabled(False)
        else:
            self.ui.viewModeButton.setText("Stick Mode")
            self.ui.stickSizeSpinBox.setEnabled(True)
            self.ui.ballSizeSpinBox.setEnabled(True)
        self.set_bonds(settings["bonds"])
        self.ui.ballSizeSpinBox.setValue(settings["ball_size"])
        self.ui.stickSizeSpinBox.setValue(settings["stick_size"])
        self.numbers = settings["atom_numbers"]
        self.ui.indexSizeSpinBox.setValue(settings["atom_numbers_size"])

        if settings["color_scheme"] == "Jmol":
            self.ui.colorSchemeSelect.setCurrentIndex(0)
        else:
            self.ui.colorSchemeSelect.setCurrentIndex(1)

    def save_settings(self) -> None:
        """Save the settings to a file."""
        save_name = self.ui.saveName.toPlainText()
        if save_name in ("", "Enter name"):
            self.ui.saveName.setPlainText("Enter name")
        else:
            settings = self.create_settings_dict()
            settings_file = f"{self.home_path}/settings/structure/{save_name}.json"
            with open(settings_file, "w") as f:
                json.dump(settings, f)
        self.update_settings_box()

    def set_cylinder_and_sphere_att(
        self,
        structure: Structure | Molecule | Crystal,
        ball_size: float,
        stick_size: float,
    ) -> None:
        """Set the sizes of the spheres and cylinders.

        :param structure: structure to set the sizes for
        :param ball_size: size of the spheres
        :param stick_size: size of the cylinders
        """
        structure.drawer.cylinder_radius = stick_size
        structure.drawer.sphere_default_radius = ball_size

        # Set the color scheme
        structure.drawer.color_scheme = self.ui.colorSchemeSelect.currentText()

        if self.stick_mode:
            structure.drawer.sphere_scale = self.ui.stickSizeSpinBox.value()
        else:
            structure.drawer.sphere_scale = self.ui.ballSizeSpinBox.value()
        structure.drawer.set_atom_colors()
        structure.drawer.set_cylinder_colors()
        structure.drawer.set_atom_scales()
        structure.drawer.set_atom_scale_matrices()
        structure.drawer.set_atom_model_matrices()

    def apply_changes(self) -> None:
        """Set the size of the cylinders."""
        structures = self.parent().structure_widget.structures
        for structure in structures:
            if self.stick_mode:
                structure.drawer.stick_mode = True

                atom_radius = 0.15
                stick_radius = atom_radius + 1e-3
            else:
                structure.drawer.stick_mode = False

                atom_radius = 1.0 / 6
                stick_radius = 0.075

            self.set_cylinder_and_sphere_att(structure, atom_radius, stick_radius)

            if self.bonds:
                structure.draw_bonds = True
                structure.drawer.cylinder_scale = self.ui.stickSizeSpinBox.value()
                structure.drawer.set_cylinder_dimensions()
                structure.drawer.set_cylinder_scale_matrices()
                structure.drawer.set_cylinder_model_matrices()
            else:
                structure.draw_bonds = False

        if structures:
            if self.numbers:
                self.parent().structure_widget.atom_indices_arrays = init_atom_number(structures[0])
                self.parent().structure_widget.number_scale = self.ui.indexSizeSpinBox.value()
            self.parent().structure_widget.show_atom_indices = self.numbers

            self.parent().structure_widget.set_vertex_attribute_objects()
        self.parent().structure_widget.update()

    def toggle_stick_mode(self) -> None:
        """Toggle between stick and ball mode."""
        if self.parent().structure_widget.structures[0].has_bonds:
            self.stick_mode = not self.stick_mode
            if self.stick_mode:
                if not self.bonds:
                    self.toggle_bonds()
                    self.bonds = True
                self.ui.viewModeButton.setText("Ball Mode")
                self.parent().ui.actionToggle_Bonds.setEnabled(False)
                self.ui.toggleBondsButton.setEnabled(False)
                self.ui.stickSizeSpinBox.setEnabled(True)
                self.ui.ballSizeSpinBox.setEnabled(False)
            else:
                self.ui.viewModeButton.setText("Stick Mode")
                self.ui.toggleBondsButton.setEnabled(True)
                self.ui.stickSizeSpinBox.setEnabled(True)
                self.ui.ballSizeSpinBox.setEnabled(True)
            self.apply_changes()

    def toggle_bonds(self) -> None:
        """Toggle bonds on and off."""
        if self.parent().structure_widget.structures[0].has_bonds:
            self.bonds = not self.bonds
            self.set_bonds(self.bonds)
            self.apply_changes()

    def toggle_numbers(self) -> None:
        """Toggle atom numbers on and off."""
        self.numbers = not self.numbers
        self.set_numbers(self.numbers)

    def set_numbers(self, numbers: bool) -> None:
        """Set numbers to True or False."""
        self.numbers = numbers
        structure = self.parent().structure_widget.structures[0]
        if self.numbers:
            if len(structure.atoms) < self.max_atoms_for_numbers:
                self.ui.toggleNumbersButton.setText("Hide Indices")
            else:
                self.numbers = False
                self.atom_indices_arrays = (np.zeros(1), np.zeros(1), np.zeros(1))
        else:
            self.ui.toggleNumbersButton.setText("Show Indices")
        self.apply_changes()

    def set_bonds(self, bonds: bool) -> None:
        """Set bonds to True or False."""
        if self.parent().structure_widget.structures and self.parent().structure_widget.structures[0].has_bonds:
            self.bonds = bonds
            if self.bonds:
                self.ui.toggleBondsButton.setText("Hide Bonds")
            else:
                self.ui.toggleBondsButton.setText("Show Bonds")
            self.apply_changes()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'measuring_tool.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QHeaderView, QLabel,
    QSizePolicy, QTableWidget, QTableWidgetItem, QWidget)

class Ui_measuring_tool(object):
    def setupUi(self, measuring_tool):
        if not measuring_tool.objectName():
            measuring_tool.setObjectName(u"measuring_tool")
        measuring_tool.resize(431, 438)
        measuring_tool.setMinimumSize(QSize(431, 438))
        measuring_tool.setMaximumSize(QSize(431, 438))
        self.info_text = QLabel(measuring_tool)
        self.info_text.setObjectName(u"info_text")
        self.info_text.setGeometry(QRect(10, 10, 371, 20))
        self.tableDistances = QTableWidget(measuring_tool)
        self.tableDistances.setObjectName(u"tableDistances")
        self.tableDistances.setGeometry(QRect(10, 280, 250, 140))
        self.labelAngles = QLabel(measuring_tool)
        self.labelAngles.setObjectName(u"labelAngles")
        self.labelAngles.setGeometry(QRect(270, 260, 58, 16))
        self.tableAngles = QTableWidget(measuring_tool)
        self.tableAngles.setObjectName(u"tableAngles")
        self.tableAngles.setGeometry(QRect(270, 280, 150, 140))
        self.tablePositions = QTableWidget(measuring_tool)
        self.tablePositions.setObjectName(u"tablePositions")
        self.tablePositions.setGeometry(QRect(10, 90, 410, 141))
        self.labelPositions = QLabel(measuring_tool)
        self.labelPositions.setObjectName(u"labelPositions")
        self.labelPositions.setGeometry(QRect(10, 70, 91, 16))
        self.labelDistances = QLabel(measuring_tool)
        self.labelDistances.setObjectName(u"labelDistances")
        self.labelDistances.setGeometry(QRect(10, 260, 81, 16))
        self.info_text_2 = QLabel(measuring_tool)
        self.info_text_2.setObjectName(u"info_text_2")
        self.info_text_2.setGeometry(QRect(10, 40, 371, 20))
        QWidget.setTabOrder(self.tablePositions, self.tableDistances)
        QWidget.setTabOrder(self.tableDistances, self.tableAngles)

        self.retranslateUi(measuring_tool)

        QMetaObject.connectSlotsByName(measuring_tool)
    # setupUi

    def retranslateUi(self, measuring_tool):
        measuring_tool.setWindowTitle(QCoreApplication.translate("measuring_tool", u"Measurement Tool", None))
        self.info_text.setText(QCoreApplication.translate("measuring_tool", u"TextLabel", None))
        self.labelAngles.setText(QCoreApplication.translate("measuring_tool", u"Angles:", None))
        self.labelPositions.setText(QCoreApplication.translate("measuring_tool", u"Atom Positions", None))
        self.labelDistances.setText(QCoreApplication.translate("measuring_tool", u"Distances:", None))
        self.info_text_2.setText(QCoreApplication.translate("measuring_tool", u"TextLabel", None))
    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'supercell_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QLabel, QSizePolicy, QSpinBox, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(257, 204)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(60, 160, 181, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.inputSupercell_a = QSpinBox(Dialog)
        self.inputSupercell_a.setObjectName(u"inputSupercell_a")
        self.inputSupercell_a.setGeometry(QRect(30, 50, 42, 26))
        self.inputSupercell_a.setMinimum(1)
        self.inputSupercell_b = QSpinBox(Dialog)
        self.inputSupercell_b.setObjectName(u"inputSupercell_b")
        self.inputSupercell_b.setGeometry(QRect(80, 50, 42, 26))
        self.inputSupercell_b.setMinimum(1)
        self.inputSupercell_c = QSpinBox(Dialog)
        self.inputSupercell_c.setObjectName(u"inputSupercell_c")
        self.inputSupercell_c.setGeometry(QRect(130, 50, 42, 26))
        self.inputSupercell_c.setMinimum(1)
        self.labelSupercell_a = QLabel(Dialog)
        self.labelSupercell_a.setObjectName(u"labelSupercell_a")
        self.labelSupercell_a.setGeometry(QRect(40, 80, 21, 16))
        self.labelSupercell_b = QLabel(Dialog)
        self.labelSupercell_b.setObjectName(u"labelSupercell_b")
        self.labelSupercell_b.setGeometry(QRect(90, 80, 21, 16))
        self.labelSupercell_c = QLabel(Dialog)
        self.labelSupercell_c.setObjectName(u"labelSupercell_c")
        self.labelSupercell_c.setGeometry(QRect(140, 80, 21, 16))
        self.labelNumAtoms = QLabel(Dialog)
        self.labelNumAtoms.setObjectName(u"labelNumAtoms")
        self.labelNumAtoms.setGeometry(QRect(20, 110, 211, 41))
        self.labelNumAtoms.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(20, 20, 171, 16))

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Change Supercell Dimensions", None))
        self.labelSupercell_a.setText(QCoreApplication.translate("Dialog", u"N<sub>a</sub>", None))
        self.labelSupercell_b.setText(QCoreApplication.translate("Dialog", u"N<sub>b</sub>", None))
        self.labelSupercell_c.setText(QCoreApplication.translate("Dialog", u"N<sub>c</sub>", None))
        self.labelNumAtoms.setText(QCoreApplication.translate("Dialog", u"TextLabel", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Enter supercell dimensions", None))
    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trajectory.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDial, QDialog, QLabel,
    QPushButton, QSizePolicy, QSlider, QWidget)

class Ui_traj_dialog(object):
    def setupUi(self, traj_dialog):
        if not traj_dialog.objectName():
            traj_dialog.setObjectName(u"traj_dialog")
        traj_dialog.resize(400, 300)
        traj_dialog.setCursor(QCursor(Qt.ArrowCursor))
        self.PrevButton = QPushButton(traj_dialog)
        self.PrevButton.setObjectName(u"PrevButton")
        self.PrevButton.setGeometry(QRect(10, 260, 50, 32))
        self.NextButton = QPushButton(traj_dialog)
        self.NextButton.setObjectName(u"NextButton")
        self.NextButton.setGeometry(QRect(60, 260, 50, 32))
        self.verticalSlider = QSlider(traj_dialog)
        self.verticalSlider.setObjectName(u"verticalSlider")
        self.verticalSlider.setGeometry(QRect(370, 60, 22, 160))
        self.verticalSlider.setOrientation(Qt.Vertical)
        self.widget = QWidget(traj_dialog)
        self.widget.setObjectName(u"widget")
        self.widget.setGeometry(QRect(19, 19, 351, 231))
        self.speedDial = QDial(traj_dialog)
        self.speedDial.setObjectName(u"speedDial")
        self.speedDial.setGeometry(QRect(350, 250, 40, 40))
        self.speedDial.setMaximum(1000)
        self.speedDial.setValue(500)
        self.speedDial.setInvertedAppearance(True)
        self.speedDial.setInvertedControls(False)
        self.playStopButton = QPushButton(traj_dialog)
        self.playStopButton.setObjectName(u"playStopButton")
        self.playStopButton.setGeometry(QRect(120, 260, 50, 32))
        self.label = QLabel(traj_dialog)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(310, 260, 41, 20))
        self.overlayButton = QPushButton(traj_dialog)
        self.overlayButton.setObjectName(u"overlayButton")
        self.overlayButton.setGeometry(QRect(180, 260, 100, 32))
        QWidget.setTabOrder(self.PrevButton, self.NextButton)
        QWidget.setTabOrder(self.NextButton, self.playStopButton)
        QWidget.setTabOrder(self.playStopButton, self.speedDial)
        QWidget.setTabOrder(self.speedDial, self.verticalSlider)

        self.retranslateUi(traj_dialog)

        QMetaObject.connectSlotsByName(traj_dialog)
    # setupUi

    def retranslateUi(self, traj_dialog):
        traj_dialog.setWindowTitle(QCoreApplication.translate("traj_dialog", u"Molecules Controller", None))
        self.PrevButton.setText(QCoreApplication.translate("traj_dialog", u"<", None))
        self.NextButton.setText(QCoreApplication.translate("traj_dialog", u">", None))
        self.playStopButton.setText(QCoreApplication.translate("traj_dialog", u"Play", None))
        self.label.setText(QCoreApplication.translate("traj_dialog", u"Speed:", None))
        self.overlayButton.setText(QCoreApplication.translate("traj_dialog", u"Show all", None))
    # retranslateUi



"""An exporter module to write chemical structures to files."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping
    from os import PathLike
    from typing import Any

    from molara.molecule.structure import Structure


__copyright__ = "Copyright 2024, Molara"


class FileExporterError(Exception):
    """Base class for errors occurring when loading molecules from file."""


class FileFormatError(FileExporterError):
    """Raised when the file format is wrong or unsupported."""


class StructureExporter(ABC):
    """Base class for structure exporters."""

    def __init__(self, path: PathLike | str) -> None:
        """Instantiate object.

        :param path: output file path
        """
        super().__init__()
        self.path = Path(path)

    @abstractmethod
    def write_structure(self, structure: Structure) -> None:
        """Write the structure in some given format into the output file.

        :param structure: Structure object to be exported to file
        """


class XyzExporter(StructureExporter):
    """Export xyz files."""

    def write_structure(self, structure: Structure) -> None:
        """Write given structure into file with xyz format.

        :param structure: Structure object to be exported to xyz-file
        """
        lines = [
            atom.symbol + "  " + rf"{atom.position[0]}  {atom.position[1]}  {atom.position[2]}"
            for atom in structure.atoms
        ]
        with open(self.path, "w") as file:
            file.write(rf"{len(structure.atoms)}" + "\n")
            file.write("This xyz file was generated automatically by Molara!\n")
            file.write("\n".join(lines))


class GeneralExporter(StructureExporter):
    """Tries to determine the file format and calls the correct exporter."""

    _IMPORTER_BY_SUFFIX: Mapping[str, Any] = {
        ".xyz": XyzExporter,
    }

    def __init__(self, path: PathLike | str) -> None:
        """Try to determine the file format and calls the correct exporter.

        :param path: output file path
        """
        super().__init__(path)

        suffix = self.path.suffix

        try:
            self._importer = self._IMPORTER_BY_SUFFIX[suffix](path)
        except KeyError:
            try:
                self._importer = XyzExporter(path)
            except FileFormatError as err:
                msg = "Could not open file."
                raise FileFormatError(msg) from err

    def write_structure(self, structure: Structure) -> None:
        """Write given structure into file with given format.

        :param Structure: structure object to be exported to file
        """
        return self._importer.write_structure(structure)



"""A Crystals Class to be able to work with several crystals."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from molara.structure.crystal import Crystal

from molara.structure.structures import Structures


class Crystals(Structures):
    """A class to store and manipulate a list of Crystals."""

    def __init__(self) -> None:
        """Initialize the Crystals Class."""
        super().__init__()
        self.energies: list = []

        # aliases for attributes and properties from Structure
        self.mols = self._structures

        # aliases for routines from Structure
        self.get_current_mol = self._get_current_structure
        self.get_mol_by_id = self._get_structure_by_id
        self.set_next_mol = self._set_next_structure
        self.set_previous_mol = self._set_previous_structure
        self.remove_crystal = self._remove_structure

    def _get_current_structure(self) -> Crystal:
        """Return the current structure."""
        return self._structures[self._structure_id]

    @property
    def num_mols(self) -> int:
        """Number of crystals."""
        return self._num_structures

    @property
    def mol_index(self) -> int:
        """Index of currently displayed molecule."""
        return self._structure_id

    def add_crystal(self, crystal: Crystal) -> None:
        """Add a new crystals to list of crystals.

        :param crystal: Crystal object to be added to Crystals object
        """
        self._add_structure(crystal)
        self.energies.append(crystal.energy)



"""Trajectory class for manipulating the Trajectory Dialog window."""

from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib as mpl
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from PySide6.QtCore import QTimer
from PySide6.QtWidgets import (
    QDialog,
    QMainWindow,
    QVBoxLayout,
)

from molara.gui.ui_trajectory import Ui_traj_dialog

if TYPE_CHECKING:
    from molara.gui.main_window import MainWindow

__copyright__ = "Copyright 2024, Molara"

mpl.use("Qt5Agg")


class MplCanvas(FigureCanvasQTAgg):
    """A class to generate a plot from matplotlib in QT."""

    def __init__(
        self,
        # This argument is necessary to be surpassed.
        parent: MainWindow | None = None,  # noqa: ARG002
        width: int = 5,
        height: int = 4,
        dpi: int = 100,
    ) -> None:
        """Initialize a Figure by generating a subplot.

        :param parent: MainWindow: The widget of the MainWindow
        :param width: int: Width of the figure
        :param height: int: Height of the figure
        :param dpi: int: MISSING INFORMATION.
        """
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = self.fig.add_subplot(111)
        super().__init__(self.fig)


class TrajectoryDialog(QDialog):
    """Dialog for manipulating appearance of trajectories."""

    def __init__(self, parent: QMainWindow | None = None) -> None:
        """Initialize the trajectory dialog.

        :param parent: parent widget (main window)
        """
        super().__init__(
            parent,
        )  # main window widget is passed as a parent, so dialog is closed if main window is closed.

        self.ui = Ui_traj_dialog()
        self.ui.setupUi(self)

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.get_next_mol)
        self.timer.setInterval(40)

        # self.ui.checkBox.stateChanged.connect(self.show_trajectory)
        self.ui.playStopButton.clicked.connect(self.show_trajectory)
        self.ui.PrevButton.clicked.connect(self.get_prev_mol)
        self.ui.NextButton.clicked.connect(self.get_next_mol)
        self.ui.overlayButton.clicked.connect(self.show_all_molecules)
        # self.ui.verticalSlider.valueChanged.connect(self.slide_molecule)
        self.ui.verticalSlider.sliderMoved.connect(self.slide_molecule)
        self.ui.speedDial.valueChanged.connect(self.change_speed)

        layout = QVBoxLayout(self.ui.widget)
        self.sc = MplCanvas(self, width=5, height=4, dpi=100)
        layout.addWidget(self.sc)
        self.ui.widget.setLayout(layout)

        self.show_all = False

    def show_trajectory(self) -> None:
        """Show the all molecules in the current Molecules class automatically."""
        if not self.parent().mols.num_mols > 1:
            return

        if self.timer.isActive():
            self.timer.stop()
            self.ui.playStopButton.setText("Play")
            return
        self.timer.start()
        self.ui.playStopButton.setText("Stop")

    def show_all_molecules(self) -> None:
        """Show all molecules in the current Molecules class automatically."""
        if not self.parent().mols.num_mols > 1:
            return
        self.show_all = not self.show_all
        if self.show_all:
            self.ui.overlayButton.setText("Show current")
            self.parent().structure_widget.set_structure(self.parent().mols.all_molecules, reset_view=False)
        else:
            self.ui.overlayButton.setText("Show all")
            self.parent().structure_widget.set_structure([self.parent().mols.get_current_mol()], reset_view=False)

    def get_next_mol(self) -> None:
        """Call molecules object to get the next molecule and update it in the GUI."""
        if not self.parent().mols.num_mols > 1:
            return

        val = self.parent().mols.mol_index
        self.ui.verticalSlider.setValue(val + 1)
        self.parent().mols.set_next_mol()
        self.update_molecule()
        if self.parent().mols.mol_index + 1 == self.parent().mols.num_mols:
            self.timer.stop()
            self.ui.playStopButton.setText("Play")

    def get_prev_mol(self) -> None:
        """Call molecules object to get the previous molecule and update it in the GUI."""
        if not self.parent().mols.num_mols > 1:
            return

        val = self.parent().mols.mol_index
        self.ui.verticalSlider.setValue(val - 1)
        self.parent().mols.set_previous_mol()
        self.update_molecule()

    def set_slider_range(self) -> None:
        """Set the slider range to the max number of molecules."""
        self.ui.verticalSlider.setRange(0, int(self.parent().mols.num_mols) - 1)

    def slide_molecule(self) -> None:
        """Update the molecule and energy plot in dependence of the slider position."""
        if not self.parent().mols.num_mols > 1:
            return

        index = self.ui.verticalSlider.sliderPosition()
        self.parent().mols.set_mol_by_id(index)
        self.update_molecule()

    def update_molecule(self) -> None:
        """Update molecule and delete old molecule."""
        self.parent().structure_widget.delete_structure()
        self.parent().structure_widget.set_structure(
            [self.parent().mols.get_current_mol()],
            reset_view=False,
        )
        self.update_energy_plot()
        if self.show_all:
            self.ui.overlayButton.setText("Show all")
            self.show_all = False

    def change_speed(self, value: int) -> None:
        """Change speed (/ time interval) of trajectory animation.

        :param value: value that is passed from the speed dial
        """
        min_interval = 1
        max_interval = 500
        self.timer.setInterval(min_interval * (max_interval / min_interval) ** (value * 0.001))

    def initial_energy_plot(self) -> None:
        """Plot the energies of the molecules in the molecules object."""
        self.sc.axes.cla()
        (self.energy_plot,) = self.sc.axes.plot(
            np.arange(self.parent().mols.num_mols),
            self.parent().mols.energies,
            "x-",
        )
        (self.current_energy_plot,) = self.sc.axes.plot(
            self.parent().mols.mol_index,
            self.parent().mols.energies[self.parent().mols.mol_index],
            "o",
        )
        self.sc.axes.set_xlabel(r"steps")
        self.sc.axes.set_ylabel(r"energy$\,/\,E_\mathrm{h}$")
        self.sc.fig.tight_layout()
        self.sc.fig.subplots_adjust(bottom=0.22, right=0.99)
        self.sc.draw()

    def update_energy_plot(self) -> None:
        """Update the energy plot, where the current structure is shown in a different color."""
        energies, mol_index = self.parent().mols.energies, self.parent().mols.mol_index
        self.current_energy_plot.set_xdata([mol_index])
        self.current_energy_plot.set_ydata([energies[mol_index]])
        self.sc.draw()

    def reset(self) -> None:
        """Clear the energy plot, update slider range."""
        self.sc.axes.cla()
        self.sc.draw()
        self.set_slider_range()



"""Contains the Atom class and a dictionary of elements."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from molara.structure.basisset import Basisset

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

__copyright__ = "Copyright 2024, Molara"

with open(Path(__file__).absolute().parent / "periodic_table.json", encoding="utf-8") as ptable_json:
    _pt_data = json.load(ptable_json)

with open(Path(__file__).absolute().parent / "atom_colors.json", encoding="utf-8") as atom_colors_json:
    _atom_colors = json.load(atom_colors_json)


class Atom:
    """Creates an Atom object."""

    def __init__(self, atomic_number: int, position: ArrayLike) -> None:
        """Create an Atom object.

        :param atomic_number: atomic number (nuclear charge number) of the atom
        :param position: cartesian coordinates of atom location
        """
        self.symbol = atomic_number_to_symbol(atomic_number)
        self.name = _pt_data[self.symbol]["Name"]
        self.atomic_number = atomic_number
        self.atomic_mass = _pt_data[self.symbol]["Atomic mass"]
        try:
            self.electronegativity = _pt_data[self.symbol]["X"]
        except KeyError:
            self.electronegativity = None
        jmol_color = (
            np.array(
                tuple(int(_atom_colors["Jmol"][self.symbol].strip("#")[i : i + 2], 16) for i in (0, 2, 4)),
                dtype=np.float64,
            )
            / 255
        )
        cpk_color = np.array(_atom_colors["CPK_ase"][self.symbol], dtype=np.float64)
        self.color = {
            "CPK": cpk_color,
            "Jmol": jmol_color,
        }
        self.vdw_radius = _pt_data[self.symbol]["Van der waals radius"]
        self.basis_set = Basisset()
        self.position = np.array([])
        self.set_position(position)

    def set_position(self, position: ArrayLike) -> None:
        """Set the position of the atom.

        :param position: new position of the atom
        :return: None
        """
        self.position = np.array(position, dtype=np.float64)


def element_symbol_to_atomic_number(symbol: str, h_isotopes: bool = False) -> int:
    """Define a dictionary mapping element symbols to atomic numbers.

    :param symbol: atomic symbol (element) of the atom
    :param h_isotopes: include hydrogen isotopes (deuterium and tritium) in the dictionary.
    :return: atomic number (nuclear charge number) of the atom
    """
    symbol_to_atomic_number = {element: _pt_data[element]["Atomic no"] for element in _pt_data}
    if not h_isotopes:
        symbol_to_atomic_number.pop("D")
        symbol_to_atomic_number.pop("T")
    return symbol_to_atomic_number.get(symbol, 0)


def atomic_number_to_symbol(atomic_number: int) -> str:
    """Define a dictionary mapping element symbols to atomic numbers.

    :param atomic_number: atomic number (nuclear charge number) of the atom
    """
    symbol_to_atomic_number = {
        _pt_data[element]["Atomic no"]: element for element in _pt_data if element not in ["D", "T"]
    }
    return symbol_to_atomic_number[atomic_number]


elements = _pt_data



"""An importer class for all read in functions."""

from __future__ import annotations

import locale
from abc import ABC, abstractmethod
from fnmatch import fnmatch
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from molara.structure.atom import element_symbol_to_atomic_number
from molara.structure.io.importer_crystal import PoscarImporter, PymatgenImporter, VasprunImporter
from molara.structure.molecule import Molecule
from molara.structure.molecules import Molecules
from molara.structure.mos import Mos

if TYPE_CHECKING:
    from collections.abc import Mapping
    from os import PathLike
    from typing import Any

    from molara.molecule.crystal import Crystal
    from molara.molecule.crystals import Crystals

    try:
        from cclib.data import ccData
    except ImportError:
        ccData = Any  # noqa: N816

__copyright__ = "Copyright 2024, Molara"

bohr_to_angstrom = 5.29177210903e-1


class FileImporterError(Exception):
    """Base class for errors occurring when loading molecules from file."""


class FileFormatError(FileImporterError):
    """Raised when the file format is wrong or unsupported."""


class MoleculesImporter(ABC):
    """Base class for importers loading molecules from files."""

    def __init__(self, path: PathLike | str) -> None:
        """Instantiate MoleculesImporter object.

        :param path: input file path
        """
        super().__init__()

        self.path = Path(path)

    @abstractmethod
    def load(self) -> Molecules | Crystals:
        """Read the file in self.path and creates a Molecules object."""


# Note: distinction between MoleculesImporter and CrystalImporter will become
# obsolete once both Molecules and Crystals have a common base class.
class CrystalImporter(ABC):
    """Base class for importers loading molecules from files."""

    def __init__(self, path: PathLike | str) -> None:
        """Instantiate CrystalImporter object.

        :param path: input file path
        """
        super().__init__()

        self.path = Path(path)

    @abstractmethod
    def load(self) -> Crystal:
        """Read the file in self.path and creates a Molecules object."""


class XyzImporter(MoleculesImporter):
    """Import xyz files."""

    def _molecule_from_xyz(self, lines: list[str]) -> Molecule:
        """Create a Molecule object from the lines of an xyz file.

        :param lines: The lines of the xyz file.
        :return: The Molecule object.
        """
        num_atoms = int(lines[0])
        atomic_numbers = []
        coordinates = []

        for line in lines[2 : 2 + num_atoms]:
            atom_info = line.split()
            if atom_info[0].isnumeric():
                atomic_numbers.append(int(atom_info[0]))
            else:
                token = atom_info[0].capitalize()
                atomic_numbers.append(element_symbol_to_atomic_number(token))
            coordinates.append([float(coord) for coord in atom_info[1:4]])

        return Molecule(np.array(atomic_numbers), np.array(coordinates), lines[1])

    def load(self) -> Molecules:
        """Read the file in self.path and creates a Molecules object."""
        molecules = Molecules()

        with open(self.path, encoding="utf-8") as file:
            lines = file.readlines()

        num_atoms = int(lines[0])
        finished = False
        max_mols = int(1e100)
        start_line = 0
        while not finished and max_mols >= molecules.num_mols:
            end_line = start_line + num_atoms + 2
            molecules.add_molecule(
                self._molecule_from_xyz(lines[start_line:end_line]),
            )
            finished = (
                start_line + 2 + num_atoms >= len(lines) or not lines[start_line + 2 + num_atoms].strip().isdigit()
            )
            start_line = end_line
        return molecules


class CoordImporter(MoleculesImporter):
    """Importer from *.coord files."""

    def load(self) -> Molecules:
        """Read the file in self.path and creates a Molecules object."""
        molecules = Molecules()

        with open(self.path, encoding=locale.getpreferredencoding(do_setlocale=False)) as file:
            lines = file.readlines()  # To skip first row

        atomic_numbers = []

        coordinates = []

        for line in lines[1:]:
            if "$" in line:
                break

            atom_info = line.split()
            if atom_info[-1].isnumeric():
                atomic_numbers.append(int(atom_info[-1]))
            else:
                atom_info[-1] = atom_info[-1].capitalize()
                atomic_numbers.append(element_symbol_to_atomic_number(atom_info[-1]))
            coordinates.append([float(coord) * 0.529177249 for coord in atom_info[:3]])

        molecules.add_molecule(
            Molecule(np.array(atomic_numbers), np.array(coordinates)),
        )

        return molecules


class MoldenImporter(MoleculesImporter):
    """Importer from *.molden files."""

    def load(self) -> Molecules:  # noqa: C901
        """Read the file in self.path and creates a Molecules object."""
        molecules = Molecules()

        with open(self.path, encoding=locale.getpreferredencoding(do_setlocale=False)) as file:
            lines = file.readlines()

        i = 0
        spherical_harmonics = ["[5D]", "[7F]", "[9G]"]

        while i < len(lines):
            if "[Atoms]" in lines[i]:
                i_start = i
                i += 1
                while "[" not in lines[i]:
                    i += 1
                atomic_numbers, coordinates = self.get_atoms(lines[i_start:i])
            if "[GTO]" in lines[i] or "[STO]" in lines[i]:
                i_start = i
                i += 1
                while "[" not in lines[i]:
                    i += 1
                basisset = self.get_basisset(lines[i_start:i])
            for sph_key in spherical_harmonics:
                if sph_key in lines[i]:
                    msg = "Spherical Harmonics not implemented."
                    raise FileFormatError(msg)
            if "[MO]" in lines[i]:
                i_start = i
                i += 1
                while "[" not in lines[i]:
                    i += 1
                    if i == len(lines):
                        break
                (
                    mo_coefficients,
                    labels,
                    energies,
                    spins,
                    occupations,
                ) = self.get_mo_coefficients(lines[i_start:i])
            i += 1

        molecules.add_molecule(
            Molecule(np.array(atomic_numbers), np.array(coordinates)),  # type: ignore[reportPossiblyUnboundVariable]
        )
        molecules.mols[0].mos = Mos(labels, energies, spins, occupations)  # type: ignore[reportPossiblyUnboundVariable]
        molecules.mols[0].mos.coefficients = np.array(mo_coefficients)  # type: ignore[reportPossiblyUnboundVariable]

        # WATCH OUT ONLY FOR GTOs!!!!!!!!
        for i, atom in enumerate(basisset):  # type: ignore[reportPossiblyUnboundVariable]
            molecules.mols[0].atoms[i].basis_set.basis_type = "GTO"
            molecules.mols[0].atoms[i].basis_set.generate_orbitals(
                atom["shells"],
                atom["exponents"],
                atom["coefficients"],
                molecules.mols[0].atoms[i].position,
            )
            molecules.mols[0].aos.extend(
                molecules.mols[0].atoms[i].basis_set.orbitals_list,
            )
        return molecules

    def get_atoms(self, lines: list[str]) -> tuple[list[int], list[list[float]]]:
        """Read the atomic numbers and coordinates from the lines of the atoms block.

        :param lines: The lines of the atom block.
        :return: The atomic numbers and coordinates.
        """
        atomic_numbers = []
        coordinates = []

        i = 0
        if "Angs" in lines[i]:
            angstrom = True
        elif "AU" in lines[i]:
            angstrom = False
        else:
            msg = "No unit specified in molden Atoms input."
            raise FileFormatError(msg)
        i += 1
        while i < len(lines):
            atom_info = lines[i].split()
            atomic_numbers.append(int(atom_info[2]))
            if not angstrom:
                coordinates.append(
                    [float(coord) * bohr_to_angstrom for coord in atom_info[3:6]],
                )
            else:
                coordinates.append([float(coord) for coord in atom_info[3:6]])
            i += 1

        return atomic_numbers, coordinates

    def get_basisset(self, lines: list[str]) -> list:  # noqa: C901
        """Read the basis set from the lines of the basisset block.

        :param lines: The lines of the basis set block.
        :return: The basis set.
        """
        i = 0
        if "STO" in lines[i]:
            msg = "STO type not implemented."
            raise FileFormatError(msg)
        i += 2
        coefficients = []
        exponents = []
        shells = ["s", "p", "d", "f", "g", "h", "i", "j", "k"]
        basisset: list = [
            {
                "shells": [],
                "exponents": [],
                "coefficients": [],
            },
        ]
        new_basisset_dict: dict = {
            "shells": [],
            "exponents": [],
            "coefficients": [],
        }
        words = lines[i].split()
        while i < len(lines):
            if not words:  # check if end of gto block
                while not words:
                    i += 1
                    if i == len(lines):
                        break
                    words = lines[i].split()
                if i == len(lines):
                    break
                basisset.append(new_basisset_dict)
                i += 1  # skip line after empty line
                words = lines[i].split()
            if words[0] == "sp":
                msg = "sp type not implemented."
                raise FileFormatError(msg)
            if words[0] in shells:
                basisset[-1]["shells"].append(words[0])
            i += 1
            words = lines[i].split()
            while words and words[0] not in shells:
                if "D" in words[0]:
                    words[0] = words[0].replace("D", "E")
                if "D" in words[1]:
                    words[1] = words[1].replace("D", "E")
                exponents.append(float(words[0]))
                coefficients.append(float(words[1]))
                i += 1
                words = lines[i].split()
            basisset[-1]["exponents"].append(exponents)
            basisset[-1]["coefficients"].append(coefficients)
            exponents = []
            coefficients = []

        return basisset

    def get_mo_coefficients(
        self,
        lines: list[str],
    ) -> tuple[list[list[float]], list[str], list[float], list[int], list[float]]:
        """Read the MO coefficients from the lines of the MO block.

        :param lines: The lines of the MO block.
        :return: The MO coefficients.
        """
        i = 1
        mo_coefficients: list = []
        labels = []
        energies = []
        spins = []
        occupations = []
        while i < len(lines):
            words = lines[i].split()
            if words[0] == "Sym=":
                labels.append(words[1])
                i += 1
                words = lines[i].split()
            if words[0] == "Ene=":
                energies.append(float(words[1]))
                i += 1
                words = lines[i].split()
            if words[0] == "Spin=":
                if words[1] == "Alpha":
                    spins.append(1)
                elif words[1] == "Beta":
                    spins.append(-1)
                i += 1
                words = lines[i].split()
            if words[0] == "Occup=":
                occupations.append(float(words[1]))
                i += 1
                words = lines[i].split()
            mo_coefficients.append([])
            while words[0] != "Sym=":
                mo_coefficients[-1].append(float(words[1]))
                i += 1
                if i == len(lines):
                    break
                words = lines[i].split()
        return mo_coefficients, labels, energies, spins, occupations


class QmImporter(MoleculesImporter):
    """importer for output files of various quantum chemistry programs."""

    def __init__(self, path: PathLike | str) -> None:
        """Instantiate QmImporter object.

        :param path: input file path
        """
        import cclib

        super().__init__(path)

        self._ccparser = cclib.io.ccopen(self.path)

        if self._ccparser is None:
            msg = "Not a QM output file."
            raise FileFormatError(msg)

    def load(self) -> Molecules:
        """Read the file in self.path and creates a Molecules object."""
        data = self._ccparser.parse()

        mols: list[Molecule] = self._get_geometries(data)
        energies = self._get_electronic_energies_in_hartree(data)

        molecules = Molecules()
        molecules.mols = mols
        molecules.energies = list(energies) if energies is not None else []

        return molecules

    def _get_electronic_energies_in_hartree(
        self,
        cclib_data: ccData,
    ) -> np.ndarray | None:
        """Convert cclib energy to hartree.

        :param cclib_data: cclib data
        """
        # conversion factor used by the cclib package
        cclib_ev_in_hartree = 27.21138505

        try:
            energy = np.array(cclib_data.scfenergies)

            energy /= cclib_ev_in_hartree

        except AttributeError:
            return None
        return energy

    def _get_geometries(self, cclib_data: ccData) -> list[Molecule]:
        """Extract geometries from cclib data.

        :param cclib_data: cclib data
        """
        try:
            atoms = cclib_data.atomnos

            mols = [Molecule(atoms, coords) for coords in cclib_data.atomcoords]

        except AttributeError as err:
            msg = "Could not read atomic coordinates."
            raise FileImporterError(msg) from err

        return mols


class GeneralImporter(MoleculesImporter):
    """Tries to determine the file format and calls the correct importer."""

    _IMPORTER_BY_SUFFIX: Mapping[str, Any] = {
        ".xyz": XyzImporter,
        ".trj": XyzImporter,
        ".log": XyzImporter,
        ".coord": CoordImporter,
        ".POSCAR": PoscarImporter,
        ".CONTCAR": PoscarImporter,
        ".vasp": PoscarImporter,
        ".cif": PymatgenImporter,
        ".xml": VasprunImporter,
        ".molden": MoldenImporter,
    }

    def __init__(
        self,
        path: PathLike | str,
    ) -> None:
        """Instantiate GeneralImporter object.

        :param path: input file path
        """
        super().__init__(path)

        suffix = self.path.suffix
        fname = self.path.name

        try:
            if suffix:
                self._importer = self._IMPORTER_BY_SUFFIX[suffix](path)
            elif fnmatch(fname, "*POSCAR*") or fnmatch(fname, "*CONTCAR*"):
                self._importer = PoscarImporter(path)
            elif fnmatch(fname, "*.json*") or fnmatch(fname, "*.mson*"):
                self._importer = PymatgenImporter(path)
        except KeyError:
            # Instead of the cclib import, we should open a dialog window to manually select the importer.
            try:
                self._importer = QmImporter(path)
            except FileFormatError as err:
                msg = "Could not open file."
                raise FileFormatError(msg) from err

    def load(self) -> Molecules | Crystals:
        """Read the file in self.path and creates a Molecules object."""
        return self._importer.load()



"""A module for the Structure class."""

from __future__ import annotations

import numpy as np
from scipy import spatial

from molara.structure.atom import Atom
from molara.structure.drawer import Drawer

__copyright__ = "Copyright 2024, Molara"

NO_BONDS = np.array([[-1, -1]], dtype=np.int_)


class Structure:
    """Base class for a structure with a set of atoms. Molecule and Crystal inherit from this."""

    def __init__(
        self: Structure,
        atomic_numbers: np.ndarray,
        coordinates: np.ndarray,
        draw_bonds: bool = True,
    ) -> None:
        """Create a new Structure object.

        :param atomic_numbers: np.ndarray: atomic numbers of a atoms
        :param coordinates: np.ndarray: coordinates of the atoms
        :param header:str: header from the imported file
        """
        self.atomic_numbers = np.array(atomic_numbers)
        self.coords = coordinates
        self.atoms = []
        self.vdw_rads: list[np.float32] = []
        self.unique_atomic_numbers: list[int] = []

        for i, atomic_number in enumerate(atomic_numbers):
            atom = Atom(atomic_number, coordinates[i])
            self.atoms.append(atom)
            if atomic_number not in self.unique_atomic_numbers:
                self.unique_atomic_numbers.append(atomic_number)

        self.molar_mass: float = np.sum([atom.atomic_mass for atom in self.atoms])

        self.bond_distance_factor = 1.0 / 1.75  # (sum of vdw radii) / 1.75 is the maximum distance for a bond
        self.draw_bonds = draw_bonds
        self.bonded_pairs = NO_BONDS
        self.bonds_calculated = False
        if self.draw_bonds:
            self.bonded_pairs = self.calculate_bonds()
            self.bonds_calculated = True

        self.drawer = Drawer(self.atoms, self.bonded_pairs, self.draw_bonds)
        self.n_at = len(self.atoms)

    def __copy__(self: Structure) -> Structure:
        """Create a copy of the structure."""
        return type(self)(
            self.atomic_numbers,
            np.array([atom.position for atom in self.atoms]),
            draw_bonds=self.draw_bonds,
        )

    def compute_collision(self: Structure, coordinate: np.ndarray) -> int | None:
        """Compute if the given coordinate is equal to the coordinate of an existing atom.

        Return None if no atom collides.

        :param coordinate: Coordinate to check whether they are equal to position of an atom
        """
        dist_threshold = 1e-10
        for i, atom in enumerate(self.atoms):
            dist = np.linalg.norm(atom.position - coordinate)
            if dist < dist_threshold:
                return i
        return None

    @property
    def center_of_mass(self: Structure) -> np.ndarray:
        """Returns the center of mass of the structure."""
        return np.average(
            [atom.position for atom in self.atoms],
            weights=[atom.atomic_mass for atom in self.atoms],
            axis=0,
        )

    def center_coordinates(self: Structure) -> None:
        """Centers the structure around the center of mass."""
        self.center = self.center_of_mass
        for _i, atom in enumerate(self.atoms):
            position = atom.position - self.center
            atom.set_position(position)

        self.drawer.update_atoms(self.atoms)
        if self.draw_bonds:
            self.drawer.update_bonds()

    def calculate_bonds(self: Structure) -> np.ndarray:
        """Calculate the bonded pairs of atoms."""
        same_position_threshold = 1e-3
        bonded_pairs = []

        vdw_radii = np.array([atom.vdw_radius for atom in self.atoms])
        coordinates = np.array([atom.position for atom in self.atoms])

        max_distance = 2.0 * vdw_radii.max() * self.bond_distance_factor
        tree = spatial.cKDTree(coordinates)

        for i, j in tree.query_pairs(max_distance):
            atom1_radius, atom2_radius = vdw_radii[i], vdw_radii[j]
            distance = np.linalg.norm(coordinates[j] - coordinates[i])

            mean_radii = (atom1_radius + atom2_radius) * self.bond_distance_factor
            if distance <= mean_radii and distance > same_position_threshold:
                bonded_pairs.append((i, j))

        if bonded_pairs:
            return np.array(bonded_pairs)

        return NO_BONDS

    @property
    def has_bonds(self) -> bool:
        """Specifies whether structure contains any bonds that could be displayed."""
        return self.bonded_pairs[0][0] != -1

    def toggle_bonds(self: Structure) -> None:
        """Toggles the bonds on and off."""
        self.draw_bonds = not self.draw_bonds
        if not self.draw_bonds:
            return
        if not self.bonds_calculated:
            self.bonded_pairs = self.calculate_bonds()
            self.bonds_calculated = True
            self.drawer.update_bonds(self.bonded_pairs, self.draw_bonds)
            return
        self.drawer.update_bonds()

    def add_atom(
        self: Structure,
        atomic_number: int,
        coordinate: np.ndarray,
    ) -> None:
        """Add an atom to the structure.

        :param atomic_number: atomic number (nuclear charge number) of the atom
        :param coordinate: cartesian coordinates of atom location
        """
        atom = Atom(atomic_number, coordinate)
        self.atoms.append(atom)
        self.bonded_pairs = self.calculate_bonds()
        self.drawer = Drawer(self.atoms, self.bonded_pairs, draw_bonds=self.draw_bonds)
        self.atomic_numbers = np.append(self.atomic_numbers, atomic_number)
        self.coords = np.append(self.coords, coordinate)
        self.n_at += 1
        self.molar_mass += atom.atomic_mass

    def remove_atom(self: Structure, index: int) -> None:
        """Remove an atom from the structure.

        :param index: list index of the atom that shall be removed
        """
        self.n_at -= 1
        self.molar_mass -= self.atoms[index].atomic_mass
        self.atoms.pop(index)

        if self.n_at != 0:
            self.bonded_pairs = self.calculate_bonds()
            self.drawer = Drawer(self.atoms, self.bonded_pairs, draw_bonds=self.draw_bonds)

        self.atomic_numbers = np.delete(self.atomic_numbers, index)
        self.coords = np.delete(self.coords, index)



"""Utility functions for testing."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import pyrr


def assert_vectors_equal(
    vector1: pyrr.Vector3 | pyrr.Quaternion | np.ndarray | pyrr.Matrix33 | pyrr.Matrix44,
    vector2: pyrr.Vector3 | pyrr.Quaternion | np.ndarray | pyrr.Matrix33 | pyrr.Matrix44,
) -> None:
    """Assert that two lists are equal."""
    np.testing.assert_allclose(
        np.array(vector1),
        np.array(vector2),
        atol=1e-7,
    )



"""Module for the Basisset class."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from molara.eval.aos import calculate_aos

if TYPE_CHECKING:
    from molara.structure.basisset import Orbital

__copyright__ = "Copyright 2024, Molara"


class Mos:
    """Class to store either an STO or GTO basisset for each atom in the same order as in molecule."""

    def __init__(
        self,
        labels: list | None = None,
        energies: list | None = None,
        spins: list | None = None,
        occupations: list | None = None,
    ) -> None:
        """Initialize the Mos class.

        :param labels: list of labels for the mos
        :param energies: list of energies for the mos
        :param spins: list of spins for the mos
        :param occupations: list of occupations for the mos
        :return:
        """
        if labels is not None:
            self.labels = labels
        else:
            self.labels = []
        if energies is not None:
            self.energies = energies
        else:
            self.energies = []
        if spins is not None:
            self.spins = spins
        else:
            self.spins = []
        if occupations is not None:
            self.occupations = occupations
        else:
            self.occupations = []
        self.coefficients: np.ndarray = np.array([])

    def calculate_mo_cartesian(  # noqa: C901
        self,
        index: int,
        aos: list[Orbital],
        electron_position: np.ndarray,
    ) -> float:
        """Calculate the value of one mo for a given electron position. Cartesian only!.

        :param index: index of the mo
        :param aos: list of all the aos
        :param electron_position: position of the electron
        :return: value of the mo
        """
        s = 0
        p = 1
        d = 2
        f = 3
        g = 4
        mo = 0
        mo_coefficients = self.coefficients[index]
        i = 0
        while i < len(mo_coefficients):
            shell = sum(aos[i].ijk)
            ao_values = calculate_aos(
                electron_position,
                aos[i].position,
                aos[i].exponents,
                aos[i].coefficients,
                shell,
            )
            if shell == s:
                mo += mo_coefficients[i] * ao_values[0]
                i += 1
            elif shell == p:
                for j in range(3):
                    mo += mo_coefficients[i] * ao_values[j]
                    i += 1
            elif shell == d:
                for j in range(6):
                    mo += mo_coefficients[i] * ao_values[j]
                    i += 1
            elif shell == f:
                for j in range(10):
                    mo += mo_coefficients[i] * ao_values[j]
                    i += 1
            elif shell == g:
                for j in range(15):
                    mo += mo_coefficients[i] * ao_values[j]
                    i += 1
            else:
                msg = f"The shell {shell} type is not supported."
                raise TypeError(msg)
        return mo



"""Dialog for specifying supercell dims."""

from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtWidgets import QDialog, QMainWindow

from molara.gui.ui_supercell_dialog import Ui_Dialog
from molara.structure.crystal import Crystal

if TYPE_CHECKING:
    intvec3 = list[int]

__copyright__ = "Copyright 2024, Molara"


class SupercellDialog(QDialog):
    """Dialog for specifying supercell dimensions.

    Supercell size given by user, object of type Crystal is instantiated
    and passed to main window"s OpenGL widget for rendering.
    """

    def __init__(self, parent: QMainWindow) -> None:
        """Create a CrystalDialog object.

        :param parent: parent widget (main window)
        """
        super().__init__(
            parent,
        )  # main window widget is passed as a parent, so dialog is closed if main window is closed.
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.main_window = parent
        # the supercell_dims attribute's purpose is to be overwritten
        # by a list that is passed by a Crystal object. The entries of this
        # list are changed in the accept routine, thus changing the original
        # list in the crystal object.
        self.supercell_dims: list[int] = [1, 1, 1]

        self.set_event_handlers()

    def get_supercell_dims(self, supercell_dims: list[int]) -> bool:
        """Open dialog for supercell size specification.

        :param supercell_dims: supercell dimensions (e.g. [2, 4, 3] for 2x4x3 supercell)
        """
        self.set_supercell_dims(supercell_dims)
        return self.exec()

    def set_event_handlers(self) -> None:
        """Set event handlers for supercell dialog."""
        self.ui.inputSupercell_a.valueChanged.connect(self.update_num_atoms_label)
        self.ui.inputSupercell_b.valueChanged.connect(self.update_num_atoms_label)
        self.ui.inputSupercell_c.valueChanged.connect(self.update_num_atoms_label)

    def set_supercell_dims(self, supercell_dims: intvec3) -> None:
        """Set supercell dimensions.

        :param supercell_dims: supercell dimensions (e.g. [2, 4, 3] for 2x4x3 supercell)
        """
        self.supercell_dims = supercell_dims
        self.ui.inputSupercell_a.setValue(supercell_dims[0])
        self.ui.inputSupercell_b.setValue(supercell_dims[1])
        self.ui.inputSupercell_c.setValue(supercell_dims[2])
        self.update_num_atoms_label()

    def update_num_atoms_label(self) -> None:
        """Update number of atoms label."""
        if not self.main_window.structure_widget.structures:
            return
        crystal = self.main_window.structure_widget.structures[0]
        if not isinstance(crystal, Crystal):
            return

        dim_a, dim_b, dim_c = (
            self.ui.inputSupercell_a.value(),
            self.ui.inputSupercell_b.value(),
            self.ui.inputSupercell_c.value(),
        )
        self.ui.labelNumAtoms.setText(
            f"Number of atoms:\n{crystal.calc_number_of_supercell_atoms([dim_a, dim_b, dim_c])}",
        )

    def accept(self) -> None:
        """Submit contents of supercell dialog.

        entries of supercell_dims are changed in order to pass
        the user-specified dimensions to the Crystal object.
        """
        dim_a, dim_b, dim_c = (
            self.ui.inputSupercell_a.value(),
            self.ui.inputSupercell_b.value(),
            self.ui.inputSupercell_c.value(),
        )
        # the purpose of the single assign of the elements is not to overwrite the list object.
        self.supercell_dims[0] = dim_a
        self.supercell_dims[1] = dim_b
        self.supercell_dims[2] = dim_c
        self.close()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'builder.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QDoubleSpinBox, QHeaderView, QLineEdit, QPushButton,
    QSizePolicy, QTableWidget, QTableWidgetItem, QTextBrowser,
    QWidget)

class Ui_builder(object):
    def setupUi(self, builder):
        if not builder.objectName():
            builder.setObjectName(u"builder")
        builder.resize(447, 310)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(builder.sizePolicy().hasHeightForWidth())
        builder.setSizePolicy(sizePolicy)
        builder.setMinimumSize(QSize(447, 310))
        builder.setMaximumSize(QSize(447, 310))
        self.buttonBox = QDialogButtonBox(builder)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(90, 270, 341, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.Box_1BondDistance = QDoubleSpinBox(builder)
        self.Box_1BondDistance.setObjectName(u"Box_1BondDistance")
        self.Box_1BondDistance.setGeometry(QRect(80, 230, 61, 31))
        self.Box_1BondDistance.setLocale(QLocale(QLocale.English, QLocale.UnitedStates))
        self.Box_1BondDistance.setMaximum(99999.000000000000000)
        self.Box_2BondAngle = QDoubleSpinBox(builder)
        self.Box_2BondAngle.setObjectName(u"Box_2BondAngle")
        self.Box_2BondAngle.setGeometry(QRect(160, 230, 61, 31))
        self.Box_2BondAngle.setLocale(QLocale(QLocale.English, QLocale.UnitedStates))
        self.Box_2BondAngle.setMaximum(360.000000000000000)
        self.Box_3DihedralAngle = QDoubleSpinBox(builder)
        self.Box_3DihedralAngle.setObjectName(u"Box_3DihedralAngle")
        self.Box_3DihedralAngle.setGeometry(QRect(240, 230, 61, 31))
        self.Box_3DihedralAngle.setLocale(QLocale(QLocale.English, QLocale.UnitedStates))
        self.Box_3DihedralAngle.setMaximum(360.000000000000000)
        self.AddAtomButton = QPushButton(builder)
        self.AddAtomButton.setObjectName(u"AddAtomButton")
        self.AddAtomButton.setGeometry(QRect(310, 230, 51, 31))
        self.Box_0Element = QLineEdit(builder)
        self.Box_0Element.setObjectName(u"Box_0Element")
        self.Box_0Element.setGeometry(QRect(20, 230, 41, 31))
        self.Box_0Element.setLocale(QLocale(QLocale.English, QLocale.UnitedStates))
        self.tableWidget = QTableWidget(builder)
        self.tableWidget.setObjectName(u"tableWidget")
        self.tableWidget.setGeometry(QRect(20, 20, 401, 161))
        self.ErrorMessageBrowser = QTextBrowser(builder)
        self.ErrorMessageBrowser.setObjectName(u"ErrorMessageBrowser")
        self.ErrorMessageBrowser.setGeometry(QRect(20, 190, 381, 21))
        self.DeleteAtomButton = QPushButton(builder)
        self.DeleteAtomButton.setObjectName(u"DeleteAtomButton")
        self.DeleteAtomButton.setGeometry(QRect(370, 230, 51, 31))
        QWidget.setTabOrder(self.tableWidget, self.ErrorMessageBrowser)
        QWidget.setTabOrder(self.ErrorMessageBrowser, self.Box_0Element)
        QWidget.setTabOrder(self.Box_0Element, self.Box_1BondDistance)
        QWidget.setTabOrder(self.Box_1BondDistance, self.Box_2BondAngle)
        QWidget.setTabOrder(self.Box_2BondAngle, self.Box_3DihedralAngle)
        QWidget.setTabOrder(self.Box_3DihedralAngle, self.AddAtomButton)
        QWidget.setTabOrder(self.AddAtomButton, self.DeleteAtomButton)

        self.retranslateUi(builder)
        self.buttonBox.accepted.connect(builder.accept)
        self.buttonBox.rejected.connect(builder.reject)

        QMetaObject.connectSlotsByName(builder)
    # setupUi

    def retranslateUi(self, builder):
        builder.setWindowTitle(QCoreApplication.translate("builder", u"Z-Matrix Builder", None))
        self.AddAtomButton.setText(QCoreApplication.translate("builder", u"Add", None))
        self.DeleteAtomButton.setText(QCoreApplication.translate("builder", u"Delete", None))
    # retranslateUi



"""Contains the main window of the application."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

from PySide6.QtCore import QCoreApplication
from PySide6.QtWidgets import QFileDialog, QMainWindow, QMessageBox

from molara.gui.builder import BuilderDialog
from molara.gui.crystal_dialog import CrystalDialog
from molara.gui.export_image_dialog import ExportImageDialog
from molara.gui.measuring_tool_dialog import MeasurementDialog
from molara.gui.structure_customizer_dialog import StructureCustomizerDialog
from molara.gui.supercell_dialog import SupercellDialog
from molara.gui.trajectory_dialog import TrajectoryDialog
from molara.gui.ui_form import Ui_MainWindow
from molara.structure.crystal import Crystal
from molara.structure.crystals import Crystals
from molara.structure.io.exporter import GeneralExporter
from molara.structure.io.importer import GeneralImporter, PoscarImporter
from molara.structure.molecules import Molecules

if TYPE_CHECKING:
    from os import PathLike

    from molara.gui.structure_widget import StructureWidget

__copyright__ = "Copyright 2024, Molara"


ENABLED, DISABLED = True, False


class MainWindow(QMainWindow):
    """Creates a MainWindow object."""

    def __init__(self, parent: QMainWindow | None = None) -> None:
        """Create a MainWindow object.

        :param parent: parent widget
        """
        super().__init__(parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.export_image_dialog = ExportImageDialog(self)

        # instantiate dialog windows, pass main window as parent.
        self.trajectory_dialog = TrajectoryDialog(self)
        self.crystal_dialog = CrystalDialog(self)
        self.measurement_dialog = MeasurementDialog(self)
        self.builder_dialog = BuilderDialog(self)
        self.supercell_dialog = SupercellDialog(self)
        self.structure_customizer_dialog = StructureCustomizerDialog(self)

        self.mols = Molecules()

        self.set_action_triggers()
        self.update_action_texts()

    @property
    def structure_widget(self) -> StructureWidget:
        """Returns the StructureWidget (openGLWidget)."""
        return self.ui.openGLWidget

    def set_action_triggers(self) -> None:
        """Connect Triggers of menu actions with the corresponding routines."""
        # Start
        self.ui.actionImport.triggered.connect(self.show_file_open_dialog)
        self.ui.actionExport.triggered.connect(self.export_structure)
        self.ui.quit.triggered.connect(self.close)

        self.ui.actionExport_Snapshot.triggered.connect(
            self.export_image_dialog.show_dialog,
        )  # self.virtual_rendering_window.draw_spheres_and_export)

        # View
        self.ui.actionReset_View.triggered.connect(self.structure_widget.reset_view)
        self.ui.actionto_x_axis.triggered.connect(lambda: self.structure_widget.set_view_to_axis("x"))
        self.ui.actionto_y_axis.triggered.connect(lambda: self.structure_widget.set_view_to_axis("y"))
        self.ui.actionto_z_axis.triggered.connect(lambda: self.structure_widget.set_view_to_axis("z"))
        self.ui.actionToggle_Axes.triggered.connect(self.structure_widget.toggle_axes)
        self.ui.actionCenter_Molecule.triggered.connect(
            self.structure_widget.center_structure,
        )
        self.ui.actionOpen_Trajectory_Dialog.triggered.connect(
            self.trajectory_dialog.show,
        )
        self.ui.actionToggle_Projection.triggered.connect(self.structure_widget.toggle_projection)
        self.ui.actionExport_CameraSettings.triggered.connect(self.export_camera_settings)
        self.ui.actionImport_CameraSettings.triggered.connect(self.import_camera_settings)
        self.ui.actionOpen_Structure_Customizer.triggered.connect(
            self.show_structure_customizer_dialog,
        )
        # tools
        self.ui.actionBuilder.triggered.connect(
            self.show_builder_dialog,
        )
        self.ui.actionMeasure.triggered.connect(
            self.show_measurement_dialog,
        )

        self.ui.actionRead_POSCAR.triggered.connect(self.show_poscar)
        self.ui.actionCreate_Lattice.triggered.connect(self.crystal_dialog.show)
        self.ui.actionSupercell.triggered.connect(self.edit_supercell_dims)
        self.ui.actionToggle_UnitCellBoundaries.triggered.connect(self.structure_widget.toggle_unit_cell_boundaries)
        self.update_action_texts()

    def show_structure_customizer_dialog(self) -> None:
        """Show the structure customizer dialog."""
        self.structure_customizer_dialog.bonds = self.structure_widget.bonds
        self.structure_customizer_dialog.set_bonds(self.structure_widget.bonds)
        self.structure_customizer_dialog.show()

    def toggle_axes(self) -> None:
        """Toggle the axes."""
        self.structure_widget.toggle_axes()
        self.update_action_texts()

    def update_action_texts(self) -> None:
        """Update the texts of the menu actions."""
        text_axes = "Hide Axes" if self.structure_widget.draw_axes else "Show Axes"
        text_projection = (
            "Perspective Projection" if self.structure_widget.orthographic_projection else "Orthographic Projection"
        )
        text_unit_cell_boundaries = (
            "Hide Unit Cell Boundaries"
            if self.structure_widget.draw_unit_cell_boundaries
            else "Show Unit Cell Boundaries"
        )
        self.ui.actionToggle_Axes.setText(QCoreApplication.translate("MainWindow", text_axes, None))
        self.ui.actionToggle_Projection.setText(QCoreApplication.translate("MainWindow", text_projection, None))
        self.ui.actionToggle_UnitCellBoundaries.setText(
            QCoreApplication.translate("MainWindow", text_unit_cell_boundaries, None),
        )

    def show_init_xyz(self) -> None:
        """Read the file from terminal arguments."""
        file_name = sys.argv[1]

        self.load_molecules(file_name)

    def show_file_open_dialog(self) -> None:
        """Select a file in the file open dialog."""
        file_name = QFileDialog.getOpenFileName(
            self,
            "Import structure from file",
            dir=".",
        )[0]
        if file_name == "":
            return
        self.load_molecules(file_name)

    def load_molecules(self, path: PathLike | str) -> None:
        """Load the molecules from path.

        :param path: input file path
        """
        importer = GeneralImporter(path)
        self.mols = importer.load()

        self.structure_widget.set_structure([self.mols.get_current_mol()])

        if self.mols.num_mols > 1:
            self.ui.actionOpen_Trajectory_Dialog.setEnabled(ENABLED)
            self.trajectory_dialog.show()
            self.trajectory_dialog.initial_energy_plot()
            self.trajectory_dialog.set_slider_range()
            return

        self.trajectory_dialog.reset()
        self.trajectory_dialog.close()
        self.ui.actionOpen_Trajectory_Dialog.setEnabled(DISABLED)

    def export_structure(self) -> None:
        """Save structure to file."""
        if not self.structure_widget.structures:
            return
        file_name = QFileDialog.getSaveFileName(
            self,
            "Export structure to file",
            f"{Path.home()}",
            "*",
        )[0]
        if file_name == "":
            return
        exporter = GeneralExporter(file_name)
        exporter.write_structure(self.structure_widget.structure)

    def export_camera_settings(self) -> None:
        """Export camera settings to .npz file."""
        file_name = QFileDialog.getSaveFileName(
            self,
            "Export camera settings to file",
            ".",
            ".json",
        )[0]
        if file_name == "":
            return
        try:
            self.structure_widget.export_camera_settings(file_name)
        except Exception as e:  # noqa: BLE001
            QMessageBox.critical(self, "Error", f"Error: {e}")

    def import_camera_settings(self) -> None:
        """Import camera settings from .npz file."""
        file_name = QFileDialog.getOpenFileName(
            self,
            "Import camera settings from file",
            ".",
            "*.json",
        )[0]
        if file_name == "":
            return
        try:
            self.structure_widget.import_camera_settings(file_name)
        except Exception as e:  # noqa: BLE001
            QMessageBox.critical(self, "Error", f"Error: {e}")

    def show_measurement_dialog(self) -> None:
        """Show the measurement dialog."""
        if len(self.structure_widget.structures) == 1:
            self.measurement_dialog.show()

    def show_builder_dialog(self) -> None:
        """Show the builder dialog."""
        self.builder_dialog.show()

    def edit_supercell_dims(self) -> bool:
        """Open dialog window to edit supercell dimensions."""
        if not isinstance(self.structure_widget.structures[0], Crystal):
            # insert error message?
            return False
        crystal = self.structure_widget.structures[0]
        supercell_dims = crystal.supercell_dims
        self.supercell_dialog.show()
        self.supercell_dialog.get_supercell_dims(supercell_dims)
        # check if supercell dimensions have successfully been passed (i.e., all are >0)
        if sum(1 for component in supercell_dims if component <= 0):
            return False
        crystal.make_supercell(supercell_dims)
        self.structure_widget.set_structure([crystal])
        return True

    def show_poscar(self) -> bool | None:
        """Read poscar file and shows the first structure in this file."""
        file_name = QFileDialog.getOpenFileName(
            self,
            caption="Open POSCAR file",
            dir=".",
            filter="POSCAR Files (*)",
        )[0]
        if file_name == "":
            return None

        importer = PoscarImporter(file_name)
        crystals = importer.load()

        if not isinstance(crystals, Crystals):
            crystal = crystals.get_current_mol()
            error_message = crystal[1]
            msg_box = QMessageBox()
            msg_box.setText(error_message)
            msg_box.exec()
            return False
        self.structure_widget.set_structure(structs=[crystals.get_current_mol()])
        return True



"""A Molecules Class to be able to work with several molecules."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molara.structure.structures import Structures

if TYPE_CHECKING:
    from molara.structure.molecule import Molecule

__copyright__ = "Copyright 2024, Molara"


class Molecules(Structures):
    """A class to store and manipulate a list of Molecules."""

    def __init__(self) -> None:
        """Initialize the Molecules Class."""
        super().__init__()
        self.energies: list = []

        # aliases for attributes and properties from Structure
        self.mols: list[Molecule] = self._structures

        # aliases for routines from Structure
        self.get_current_mol = self._get_current_structure
        self.get_mol_by_id = self._get_structure_by_id
        self.set_next_mol = self._set_next_structure
        self.set_previous_mol = self._set_previous_structure
        self.set_mol_by_id = self._set_structure_by_id

    @property
    def num_mols(self) -> int:
        """Number of molecules."""
        return self._num_structures

    @property
    def all_molecules(self) -> list[Molecule]:
        """Return all molecules."""
        return self.mols

    @property
    def mol_index(self) -> int:
        """Index of currently displayed molecule."""
        return self._structure_id

    def add_molecule(self, mol: Molecule) -> None:
        """Add a new molecules to list of molecules.

        :param mol: Molecule object to be added
        """
        self._add_structure(mol)
        self.energies.append(mol.energy)

    def remove_molecule(self, mol_id: int) -> None:
        """Remove a molecule from list of molecules.

        :param mol_id: Index of the molecule
        """
        self._remove_structure(mol_id)
        self.energies.pop(mol_id)



"""Virtual (i.e., hidden) window for rendering a structure for snapshot export."""

from __future__ import annotations

from PySide6.QtWidgets import QMainWindow, QWidget

from molara.gui.ui_virtual_rendering_window import Ui_Form


class VirtualRenderingWindow(QWidget):
    """Virtual (i.e., hidden) window for rendering a structure for snapshot export."""

    def __init__(self, parent: QMainWindow) -> None:
        """Instantiate the virtual rendering window object."""
        super().__init__(
            parent,
        )
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.main_window = parent
        self.structure_widget = self.ui.openGLWidget

    def export_image(self, filename: str, custom_geometry: tuple[int, int]) -> None:
        """Export the image of the structure to the given filename.

        :param filename: the filename for the exported image
        :param custom_geometry: custom geometry (width, height) for the exported image
        """
        self.structure_widget.grabFramebuffer().save(filename)

        # rendering part
        main_structure_widget = self.main_window.structure_widget
        virtual_structure_widget = self.structure_widget

        # the window must be shown to render (apparently) but can immediately be hidden again
        self.show()
        virtual_structure_widget.adopt_config(main_structure_widget, custom_geometry)
        self.hide()
        virtual_structure_widget.update()
        virtual_structure_widget.grabFramebuffer().save(filename)



"""A Structures Class to be able to work with several structures."""

from __future__ import annotations

from molara.structure.structure import Structure

__copyright__ = "Copyright 2024, Molara"


class Structures:
    """A class to store and manipulate a list of Structures."""

    def __init__(self) -> None:
        """Initialize the Structures Class."""
        self._structures: list = []
        self._structure_id = 0

    @property
    def _num_structures(self) -> int:
        """Number of structures."""
        return len(self._structures)

    def _get_current_structure(self) -> Structure:
        """Return the current structure."""
        return self._structures[self._structure_id]

    def _get_structure_by_id(self, structure_id: int) -> Structure:
        """Return a structure of the list of structure by a given index.

        :param structure_id: list index of the structure that shall be returned
        """
        self._structure_id = structure_id
        return self._structures[self._structure_id]

    def _set_next_structure(self) -> None:
        """Return the next structure in the list of structure."""
        self._structure_id += 1
        self._structure_id %= self._num_structures

    def _set_previous_structure(self) -> None:
        """Return the previous structure of the list of structure."""
        self._structure_id -= 1
        self._structure_id %= self._num_structures

    def _set_structure_by_id(self, structure_id: int) -> None:
        """Set the structure by a given index.

        :param structure_id: list index of the structure that shall be set as current
        """
        self._structure_id = structure_id
        self._structure_id %= self._num_structures

    def _add_structure(self, struct: Structure) -> None:
        """Add a structure to the list of structures.

        :param struct: Structure object to be added to the list
        """
        if not isinstance(struct, Structure):
            msg = "The given structure is not a Structure object."
            raise TypeError(msg)
        self._structures.append(struct)

    def _remove_structure(self, structure_id: int) -> None:
        """Remove a structure from the list of structures.

        :param structure_id: list index of the structure that shall be removed
        """
        self._structures.pop(structure_id)



"""Serves the calculation of atomic orbitals."""

from __future__ import annotations

import numpy as np

__copyright__ = "Copyright 2024, Molara"


def calculate_aos(  # noqa: C901 PLR0915
    electron_coords: np.ndarray,
    atom_coords: np.ndarray,
    exponents: np.ndarray,
    coefficients: np.ndarray,
    orbital: int,
) -> np.ndarray:
    """Calculate the atomic orbitals for a given atom (cartesian).

    If the orbital is an s orbital, the function the return has size 1, if the orbital is a d orbital, the return
    ha size 6 (dxx, dyy, dzz, etc.).
    :param electron_coords: list of coordinates
    :param atom_coords: list of atom coordinates
    :param exponents: list of exponents
    :param coefficients: list of coefficients
    :param orbital: Name of the orbital (ie, s, p, d, etc.)
    :return: Value of the atomic orbital(s) for a given atom position and electron position
    """
    sqr3 = 1.73205080756887729
    sqr5 = 2.236067977499789696
    sqr7 = 2.645751311064591

    s = 0
    p = 1
    d = 2
    f = 3
    g = 4

    fxxx = 0
    fyyy = 1
    fzzz = 2
    fxyy = 3
    fxxy = 4
    fxxz = 5
    fxzz = 6
    fyzz = 7
    fyyz = 8
    fxyz = 9

    gxxxx = 0
    gyyyy = 1
    gzzzz = 2
    gxxxy = 3
    gxxxz = 4
    gyyyx = 5
    gyyyz = 6
    gzzzx = 7
    gzzzy = 8
    gxxyy = 9
    gxxzz = 10
    gyyzz = 11
    gxxyz = 12
    gyyxz = 13
    gzzxy = 14

    ngto = len(exponents)
    relative_coords = electron_coords - atom_coords
    rr = np.linalg.norm(relative_coords)
    r2 = rr * rr
    if orbital == s:
        uao = np.zeros(1)
        for ic in range(ngto):
            u = coefficients[ic] * np.exp(-exponents[ic] * r2)
            uao[0] = uao[0] + u
    elif orbital == p:
        uao = np.zeros(3)
        for ic in range(ngto):
            u = coefficients[ic] * np.exp(-exponents[ic] * r2)
            dx = relative_coords[0]
            dy = relative_coords[1]
            dz = relative_coords[2]
            uao[0] = uao[0] + dx * u
            uao[1] = uao[1] + dy * u
            uao[2] = uao[2] + dz * u
    elif orbital == d:
        uao = np.zeros(6)
        for ic in range(ngto):
            u = coefficients[ic] * np.exp(-exponents[ic] * r2)
            dx = relative_coords[0]
            dx2 = dx * dx
            dy = relative_coords[1]
            dy2 = dy * dy
            dz = relative_coords[2]
            dz2 = dz * dz
            uao[0] = uao[0] + dx2 * u
            uao[1] = uao[1] + dy2 * u
            uao[2] = uao[2] + dz2 * u
            u = sqr3 * u
            uao[3] = uao[3] + dx * dy * u
            uao[4] = uao[4] + dx * dz * u
            uao[5] = uao[5] + dy * dz * u
    elif orbital == f:
        uao = np.zeros(10)
        for ic in range(ngto):
            u = coefficients[ic] * np.exp(-exponents[ic] * r2)
            dx = relative_coords[0]
            dx2 = dx * dx
            dy = relative_coords[1]
            dy2 = dy * dy
            dz = relative_coords[2]
            dz2 = dz * dz
            dxyz = dx * dy * dz
            uao[fxxx] = uao[fxxx] + dx2 * dx * u
            uao[fyyy] = uao[fyyy] + dy2 * dy * u
            uao[fzzz] = uao[fzzz] + dz2 * dz * u
            u = sqr5 * u
            uao[fxxy] = uao[fxxy] + dx2 * dy * u
            uao[fxxz] = uao[fxxz] + dx2 * dz * u
            uao[fxyy] = uao[fxyy] + dy2 * dx * u
            uao[fyyz] = uao[fyyz] + dy2 * dz * u
            uao[fxzz] = uao[fxzz] + dz2 * dx * u
            uao[fyzz] = uao[fyzz] + dz2 * dy * u
            u = sqr3 * u
            uao[fxyz] = uao[fxyz] + dxyz * u
    elif orbital == g:
        uao = np.zeros(15)
        for ic in range(ngto):
            u = coefficients[ic] * np.exp(-exponents[ic] * r2)
            dx = relative_coords[0]
            dx2 = dx * dx
            dy = relative_coords[1]
            dy2 = dy * dy
            dz = relative_coords[2]
            dz2 = dz * dz
            dxyz = dx * dy * dz
            uao[gxxxx] = uao[gxxxx] + dx2 * dx2 * u
            uao[gyyyy] = uao[gyyyy] + dy2 * dy2 * u
            uao[gzzzz] = uao[gzzzz] + dz2 * dz2 * u
            u = sqr7 * u
            uao[gxxxy] = uao[gxxxy] + dx2 * dx * dy * u
            uao[gxxxz] = uao[gxxxz] + dx2 * dx * dz * u
            uao[gyyyx] = uao[gyyyx] + dy2 * dy * dx * u
            uao[gyyyz] = uao[gyyyz] + dy2 * dy * dz * u
            uao[gzzzx] = uao[gzzzx] + dz2 * dz * dx * u
            uao[gzzzy] = uao[gzzzy] + dz2 * dz * dy * u
            u = sqr5 / sqr3 * u
            uao[gxxyy] = uao[gxxyy] + dx2 * dy2 * u
            uao[gxxzz] = uao[gxxzz] + dx2 * dz2 * u
            uao[gyyzz] = uao[gyyzz] + dy2 * dz2 * u
            u = sqr3 * u
            uao[gxxyz] = uao[gxxyz] + dx * dxyz * u
            uao[gyyxz] = uao[gyyxz] + dy * dxyz * u
            uao[gzzxy] = uao[gzzxy] + dz * dxyz * u
    else:
        msg = "(calculate_aos): wrong GTO"
        raise TypeError(msg)
    return uao



"""File used to compile the cython modules.

It needs to be called manually with python3 cython_build.py build_ext --inplace.
"""

from __future__ import annotations

import numpy as np
from Cython.Build import cythonize
from setuptools import setup

__copyright__ = "Copyright 2024, Molara"

# Modules to be compiled and include_dirs when necessary
extensions = [
    "src/molara/rendering/cylinder.pyx",
    "src/molara/rendering/sphere.pyx",
    "src/molara/rendering/matrices.pyx",
    "src/molara/tools/mathtools.pyx",
    "src/molara/tools/raycasting.pyx",
]


# This is the function that is executed
setup(
    name="Molara",  # Required
    # A list of compiler Directives is available at
    # https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#compiler-directives
    # external to be compiled
    ext_modules=cythonize(extensions, compiler_directives={"language_level": 3}),
    include_dirs=[np.get_include()],
)
